from typing import TypeAlias

import pydantic_extra_types.color
from typing_extensions import override

from svglab import serialize


class Color(
    pydantic_extra_types.color.Color, serialize.CustomSerializable
):
    @override
    def serialize(self) -> str:
        formatter = serialize.get_current_formatter()
        result: str

        match formatter.color_mode:
            case "original":
                original = self.original()

                if isinstance(original, str):
                    result = original
                else:
                    result = self.as_named(fallback=True)
            case "auto":
                result = super().__str__()
            case "named":
                result = self.as_named(fallback=True)
            case "hex-short":
                result = self.as_hex(format="short")
            case "hex-long":
                result = self.as_hex(format="long")
            case "rgb":
                result = self.as_rgb()
            case "hsl":
                result = self.as_hsl()

        return result

    @override
    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.original()!r})"


ColorType: TypeAlias = Color



from __future__ import annotations

from typing import Annotated, Literal, TypeAlias

import lark
import pydantic
from typing_extensions import override

from svglab import serialize
from svglab.attrparse import utils


LengthUnit: TypeAlias = Literal[
    "em", "ex", "px", "in", "cm", "mm", "pt", "pc", "%"
]


@pydantic.dataclasses.dataclass
class Length(serialize.CustomSerializable):
    """Represents the SVG `<length>` type.

    A length is a number optionally followed by a unit. Available units are:

    - `em`: the font size of the element
    - `ex`: the x-height of the element's font
    - `px`: pixels
    - `in`: inches
    - `cm`: centimeters
    - `mm`: millimeters
    - `pt`: points
    - `pc`: picas
    - `%`: percentage of another value
    """

    value: float
    unit: LengthUnit | None = None

    @override
    def serialize(self) -> str:
        value = serialize.serialize(self.value)
        return f"{value}{self.unit or ''}"


@lark.v_args(inline=True)
class _Transformer(lark.Transformer[object, Length]):
    number = float
    length = Length


LengthType: TypeAlias = Annotated[
    Length,
    utils.get_validator(grammar="length.lark", transformer=_Transformer()),
]



from typing import Literal

from svglab import models
from svglab.attrs import types


class Attr(models.BaseModel):
    pass


# definitions of attributes that are both regular and presentation attributes
# are merged together; this is a bit of a hack
# TODO: figure out how the standard intends us to handle this


class FontWeight(Attr):
    font_weight: models.Attr[
        types.All
        | models.List[
            Literal[
                "normal",
                "bold",
                "bolder",
                "lighter",
                100,
                200,
                300,
                400,
                500,
                600,
                700,
                800,
                900,
            ]
        ]
        | types.Inherit
    ] = None


class FontStretch(Attr):
    font_stretch: models.Attr[
        types.All
        | Literal[
            "condensed ",
            "condensed",
            "expanded",
            "extra-condensed",
            "extra-expanded",
            "narrower",
            "normal",
            "semi-condensed",
            "semi-expanded",
            "ultra-condensed",
            "ultra-expanded",
            "wider",
        ]
        | types.Inherit
    ] = None


class FontStyle(Attr):
    font_style: models.Attr[
        types.All
        | models.List[Literal["normal", "italic", "oblique"]]
        | types.Inherit
    ] = None


class FontVariant(Attr):
    font_variant: models.Attr[
        Literal["normal", "small-caps"] | types.Inherit
    ] = None


class Fill(Attr):
    fill: models.Attr[types.Paint | Literal["freeze", "remove"]] = None



from svglab.attrs import common, presentation, regular


class Core(regular.Id, regular.XmlBase, regular.XmlLang, regular.XmlSpace):
    pass


class Presentation(
    common.Fill,
    common.FontStretch,
    common.FontStyle,
    common.FontVariant,
    common.FontWeight,
    presentation.BaselineShift,
    presentation.ClipRule,
    presentation.Color,
    presentation.ColorInterpolation,
    presentation.ColorInterpolationFilters,
    presentation.ColorProfile,
    presentation.ColorRendering,
    presentation.Cursor,
    presentation.Direction,
    presentation.Display,
    presentation.DominantBaseline,
    presentation.FillOpacity,
    presentation.FillRule,
    presentation.FontSize,
    presentation.FontSizeAdjust,
    presentation.GlyphOrientationHorizontal,
    presentation.GlyphOrientationVertical,
    presentation.ImageRendering,
    presentation.Kerning,
    presentation.LetterSpacing,
    presentation.MarkerEnd,
    presentation.MarkerMid,
    presentation.MarkerStart,
    presentation.Mask,
    presentation.Opacity,
    presentation.Overflow,
    presentation.PointerEvents,
    presentation.ShapeRendering,
    presentation.StopColor,
    presentation.StopOpacity,
    presentation.Stroke,
    presentation.StrokeDasharray,
    presentation.StrokeDashoffset,
    presentation.StrokeLinecap,
    presentation.StrokeLinejoin,
    presentation.StrokeMiterlimit,
    presentation.StrokeOpacity,
    presentation.StrokeWidth,
    presentation.TextAnchor,
    presentation.TextDecoration,
    presentation.TextRendering,
    presentation.UnicodeBidi,
    presentation.Visibility,
    presentation.WordSpacing,
    presentation.WritingMode,
):
    pass


class AnimationEvents(
    regular.OnBegin, regular.OnEnd, regular.OnLoad, regular.OnRepeat
):
    pass


class AnimationAttributeTarget(
    regular.AttributeName, regular.AttributeType
):
    pass


class AnimationTiming(
    regular.Begin,
    regular.Dur,
    regular.End,
    common.Fill,
    regular.Max,
    regular.Min,
    regular.RepeatCount,
    regular.RepeatDur,
    regular.Restart,
):
    pass


class AnimationValue(
    regular.By,
    regular.CalcMode,
    regular.From,
    regular.KeySplines,
    regular.KeyTimes,
    regular.To,
    regular.ValuesList,
):
    pass


class AnimationAddition(regular.Accumulate, regular.Additive):
    pass


class ConditionalProcessing(
    regular.RequiredFeatures,
    regular.RequiredExtensions,
    regular.SystemLanguage,
):
    pass


class DocumentEvents(
    regular.OnAbort,
    regular.OnError,
    regular.OnResize,
    regular.OnScroll,
    regular.OnUnload,
    regular.OnZoom,
):
    pass


class GraphicalEvents(
    regular.OnActivate,
    regular.OnClick,
    regular.OnFocusIn,
    regular.OnFocusOut,
    regular.OnLoad,
    regular.OnMouseDown,
    regular.OnMouseMove,
    regular.OnMouseOut,
    regular.OnMouseOver,
    regular.OnMouseUp,
):
    pass


class Event(GraphicalEvents, AnimationEvents, DocumentEvents):
    pass


class FilterPrimitives(
    regular.XCoordinate,
    regular.YCoordinate,
    regular.Width,
    regular.Height,
    regular.Result,
):
    pass


class Xlink(
    regular.XlinkActuateOnLoad,
    regular.XlinkArcrole,
    regular.XlinkHref,
    regular.XlinkRole,
    regular.XlinkShow,
    regular.XlinkTitle,
    regular.XlinkType,
):
    pass


class TransferFunction(
    regular.Amplitude,
    regular.Exponent,
    regular.Intercept,
    regular.OffsetNumber,
    regular.Slope,
    regular.TableValues,
    regular.TypeFeFunc,
):
    pass



from typing import Annotated, TypeAlias, cast, final, overload

import lark
import pydantic
from typing_extensions import Self, override

from svglab import serialize
from svglab.attrparse import utils


@final
@pydantic.dataclasses.dataclass
class Translate(serialize.CustomSerializable):
    x: float
    y: float | None = None

    @override
    def serialize(self) -> str:
        x = serialize.serialize(self.x)

        if self.y is None:
            return f"translate({x})"

        y = serialize.serialize(self.y)

        return f"translate({x}, {y})"


@final
@pydantic.dataclasses.dataclass
class Scale(serialize.CustomSerializable):
    x: float
    y: float | None = None

    @override
    def serialize(self) -> str:
        x = serialize.serialize(self.x)

        if self.y is None:
            return f"scale({x})"

        y = serialize.serialize(self.y)

        return f"scale({x}, {y})"


@pydantic.dataclasses.dataclass
class _Rotate(serialize.CustomSerializable):
    angle: float
    cx: float | None
    cy: float | None

    @pydantic.model_validator(mode="after")
    def __check_cx_cy(self) -> Self:  # pyright: ignore[reportUnusedFunction]
        cx_is_none = self.cx is None
        cy_is_none = self.cy is None

        if cx_is_none != cy_is_none:
            raise ValueError(
                "Both cx and cy must either be provided or omitted"
            )

        return self

    @override
    def serialize(self) -> str:
        angle = serialize.serialize(self.angle)

        if self.cx is None:
            return f"rotate({angle})"

        cx, cy = serialize.serialize(self.cx, cast(float, self.cy))

        return f"rotate({angle} {cx} {cy})"


class Rotate(_Rotate):
    @overload
    def __init__(self, angle: float, /) -> None: ...

    @overload
    def __init__(self, angle: float, /, cx: float, cy: float) -> None: ...

    def __init__(
        self,
        angle: float,
        /,
        cx: float | None = None,
        cy: float | None = None,
    ) -> None:
        super().__init__(angle=angle, cx=cx, cy=cy)


@final
@pydantic.dataclasses.dataclass
class SkewX(serialize.CustomSerializable):
    angle: float

    @override
    def serialize(self) -> str:
        angle = serialize.serialize(self.angle)

        return f"skewX({angle})"


@final
@pydantic.dataclasses.dataclass
class SkewY(serialize.CustomSerializable):
    angle: float

    @override
    def serialize(self) -> str:
        angle = serialize.serialize(self.angle)

        return f"skewY({angle})"


@final
@pydantic.dataclasses.dataclass
class Matrix(serialize.CustomSerializable):
    a: float
    b: float
    c: float
    d: float
    e: float
    f: float

    @override
    def serialize(self) -> str:
        a, b, c, d, e, f = serialize.serialize(
            self.a, self.b, self.c, self.d, self.e, self.f
        )

        return f"matrix({a} {b} {c} {d} {e} {f})"


TransformAction: TypeAlias = (
    Translate | Scale | Rotate | SkewX | SkewY | Matrix
)

Transform: TypeAlias = list[TransformAction]


@lark.v_args(inline=True)
class _Transformer(lark.Transformer[object, Transform]):
    number = float

    translate = Translate
    scale = Scale
    rotate = Rotate
    skew_x = SkewX
    skew_y = SkewY
    matrix = Matrix

    transform_ = utils.v_args_to_list


TransformType: TypeAlias = Annotated[
    Transform,
    utils.get_validator(
        grammar="transform.lark", transformer=_Transformer()
    ),
]



import typing
from typing import Final, Literal, TypeAlias


Xmlns: TypeAlias = Literal["http://www.w3.org/2000/svg"]
"""Type representing valid values for the `xmlns` attribute."""

DEFAULT_XMLNS: Final[Xmlns] = typing.get_args(Xmlns)[0]
"""The default value for the `xmlns` attribute."""



from svglab.attrparse.angle import Angle, AngleType, AngleUnit
from svglab.attrparse.color import Color, ColorType
from svglab.attrparse.d import (
    ArcTo,
    CubicBezierTo,
    D,
    DType,
    LineTo,
    PathCommand,
    QuadraticBezierTo,
)
from svglab.attrparse.length import Length, LengthType, LengthUnit
from svglab.attrparse.point import Point, PointType
from svglab.attrparse.points import Points, PointsType
from svglab.attrparse.transform import (
    Matrix,
    Rotate,
    Scale,
    SkewX,
    SkewY,
    Transform,
    TransformAction,
    TransformType,
    Translate,
)


__all__ = [
    "Angle",
    "AngleType",
    "AngleUnit",
    "ArcTo",
    "Color",
    "ColorType",
    "CubicBezierTo",
    "D",
    "DType",
    "Length",
    "LengthType",
    "LengthUnit",
    "LineTo",
    "Matrix",
    "PathCommand",
    "Point",
    "PointType",
    "Points",
    "PointsType",
    "QuadraticBezierTo",
    "Rotate",
    "Scale",
    "SkewX",
    "SkewY",
    "Transform",
    "TransformAction",
    "TransformType",
    "Translate",
]



import collections
import functools
from collections.abc import Generator, Iterable
from typing import Protocol, TypeVar, runtime_checkable

import bs4
from useful_types import SupportsRichComparison


_T = TypeVar("_T")
_T_cmp = TypeVar("_T_cmp", bound=SupportsRichComparison)

_AnyStr_contra = TypeVar("_AnyStr_contra", str, bytes, contravariant=True)
_AnyStr_co = TypeVar("_AnyStr_co", str, bytes, covariant=True)


@runtime_checkable
class SupportsRead(Protocol[_AnyStr_co]):
    """Protocol for objects that support reading.

    This exists because using `SupportsRead` from `typeshed` causes problems.

    Example:
    >>> from io import StringIO
    >>> buf = StringIO()
    >>> isinstance(buf, SupportsRead)
    True

    """

    def read(self, size: int | None = None, /) -> _AnyStr_co: ...


@runtime_checkable
class SupportsWrite(Protocol[_AnyStr_contra]):
    """Protocol for objects that support writing.

    This exists because using `SupportsWrite` from `typeshed` causes problems.

    Example:
    >>> from io import StringIO
    >>> buf = StringIO()
    >>> isinstance(buf, SupportsWrite)
    True

    """

    def write(self, data: _AnyStr_contra, /) -> int: ...


def is_empty(iterable: Iterable[object], /) -> bool:
    """Determine whether an iterable is empty.

    Args:
        iterable: The iterable to check.

    Returns:
        `True` if the iterable is empty, `False` otherwise.

    Examples:
        >>> is_empty([])
        True
        >>> is_empty([1, 2, 3])
        False
        >>> is_empty(range(0))
        True
        >>> is_empty(range(3))
        False

    """
    for _ in iterable:
        return False

    return True


def length(iterable: Iterable[_T], /) -> int:
    """Count the number of items in an iterable.

    Args:
        iterable: The iterable to count.

    Returns:
        The number of items in the iterable.

    Examples:
        >>> length([])
        0
        >>> length([1, 2, 3])
        3
        >>> length(range(0))
        0
        >>> length(range(3))
        3

    """
    return sum(1 for _ in iterable)


def take_last(iterable: Iterable[_T], /) -> _T | None:
    """Get the last item in an iterable.

    Args:
        iterable: The iterable to extract the last item from.

    Returns:
        The last item in the iterable, or `None` if the iterable is empty.

    Examples:
        >>> take_last([]) is None
        True
        >>> take_last([1, 2, 3])
        3
        >>> take_last(range(0)) is None
        True
        >>> take_last(range(3))
        2

    """
    return functools.reduce(lambda _, s: s, iterable, None)


def make_soup(element: bs4.PageElement, /) -> bs4.BeautifulSoup:
    soup = bs4.BeautifulSoup()
    soup.append(element)
    return soup


def get_formatter(*, indent: int) -> bs4.formatter.Formatter:
    if indent < 0:
        raise ValueError("Indent must be a non-negative integer.")

    return bs4.formatter.XMLFormatter(indent=indent)


def beautifulsoup_to_str(
    element: bs4.PageElement, /, *, pretty: bool, indent: int
) -> str:
    result: str

    match element, pretty:
        case bs4.NavigableString(), _:
            result = str(make_soup(element))
        case bs4.Tag(), True:
            formatter = get_formatter(indent=indent)
            soup = make_soup(element)

            result = soup.prettify(formatter=formatter)
        case bs4.Tag(), False:
            result = str(element)
        case _:
            msg = f"Unsupported type: {type(element)}"
            raise TypeError(msg)

    return result.strip()


def clamp(
    value: _T_cmp, /, *, min_value: _T_cmp, max_value: _T_cmp
) -> _T_cmp:
    """Clamp a value between two bounds.

    Args:
        value: The value to clamp.
        min_value: The minimum value.
        max_value: The maximum value.

    Returns:
        The clamped value. If `value` is less than `min_value`, `min_value` is
        returned. If `value` is greater than `max_value`, `max_value` is
        returned. Otherwise, `value` is returned.

    Examples:
        >>> clamp(5, min_value=0, max_value=10)
        5
        >>> clamp(-5, min_value=0, max_value=10)
        0
        >>> clamp(15, min_value=0, max_value=10)
        10

    """
    return max(min(value, max_value), min_value)


def get_all_subclasses(
    cls: type[_T], /
) -> Generator[type[_T], None, None]:
    """Recursively obtain all subclasses of a class.

    Args:
        cls: The class to obtain subclasses for.

    Yields:
        Subclasses of the given class, including subclasses of subclasses
        (and so on).

    Examples:
        >>> class A:
        ...     pass
        >>> class B(A):
        ...     pass
        >>> class C(B):
        ...     pass
        >>> list(cls.__name__ for cls in get_all_subclasses(A))
        ['B', 'C']

    """
    queue = collections.deque([cls])

    while queue:
        subclass = queue.popleft()

        if subclass is not cls:
            yield subclass

        queue.extend(subclass.__subclasses__())



import typing
from typing import Final, Literal, TypeAlias

import bidict


TagName: TypeAlias = Literal[
    "a",
    "altGlyph",
    "altGlyphDef",
    "altGlyphItem",
    "animate",
    "animateColor",
    "animateMotion",
    "animateTransform",
    "circle",
    "clipPath",
    "color-profile",
    "cursor",
    "defs",
    "desc",
    "ellipse",
    "feBlend",
    "feColorMatrix",
    "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    "feDiffuseLighting",
    "feDisplacementMap",
    "feDistantLight",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feGaussianBlur",
    "feImage",
    "feMerge",
    "feMergeNode",
    "feMorphology",
    "feOffset",
    "fePointLight",
    "feSpecularLighting",
    "feSpotLight",
    "feTile",
    "feTurbulence",
    "filter",
    "font-face-format",
    "font-face-name",
    "font-face-src",
    "font-face-uri",
    "font-face",
    "font",
    "foreignObject",
    "g",
    "glyph",
    "glyphRef",
    "hkern",
    "image",
    "line",
    "linearGradient",
    "marker",
    "mask",
    "metadata",
    "missing-glyph",
    "mpath",
    "path",
    "pattern",
    "polygon",
    "polyline",
    "radialGradient",
    "rect",
    "script",
    "set",
    "stop",
    "style",
    "svg",
    "switch",
    "symbol",
    "text",
    "textPath",
    "title",
    "tref",
    "tspan",
    "use",
    "view",
    "vkern",
]
"""Type for all SVG tag names."""


TAG_NAMES: Final[frozenset[TagName]] = frozenset(typing.get_args(TagName))
"""A set of all SVG tag names."""


def _normalize_tag_name(name: TagName, /) -> str:
    """Convert an SVG tag name to an appropriate class name.

    Args:
        name: The tag name to normalize.

    Returns:
        The normalized tag name in `PascalCase`.

    Raises:
        ValueError: If the tag name cannot be normalized.

    Examples:
    >>> _normalize_tag_name("circle")
    'Circle'
    >>> _normalize_tag_name("feGaussianBlur")
    'FeGaussianBlur'
    >>> _normalize_tag_name("font-face-name")
    'FontFaceName'

    """
    return "".join(part[0].upper() + part[1:] for part in name.split("-"))


TAG_NAME_TO_NORMALIZED: Final = bidict.frozenbidict[TagName, str](
    {tag: _normalize_tag_name(tag) for tag in TAG_NAMES}
)
"""
A bidirectional mapping from SVG tag names to normalized Python identifiers.
"""



from typing import Literal

from svglab import models
from svglab.attrs import common, types


class AlignmentBaseline(common.Attr):
    alignment_baseline: models.Attr[
        types.Auto
        | types.Inherit
        | Literal[
            "baseline",
            "before-edge",
            "text-before-edge",
            "middle",
            "central",
            "after-edge",
            "text-after-edge",
            "ideographic",
            "alphabetic",
            "hanging",
            "mathematical",
        ]
    ] = None


class BaselineShift(common.Attr):
    baseline_shift: models.Attr[
        Literal["baseline", "sub", "super"]
        | types.Percentage
        | types.Length
        | types.Inherit
    ] = None


class ClipPath(common.Attr):
    clip_path: models.Attr[types.FuncIri | types.None_ | types.Inherit] = (
        None
    )


class ClipRule(common.Attr):
    clip_rule: models.Attr[
        Literal["nonzero", "evenodd"] | types.Inherit
    ] = None


class Clip(common.Attr):
    clip: models.Attr[types.Shape | types.Auto | types.Inherit] = None


class ColorInterpolationFilters(common.Attr):
    color_interpolation_filters: models.Attr[
        types.Auto | Literal["sRGB", "linearRGB"] | types.Inherit
    ] = None


class ColorInterpolation(common.Attr):
    color_interpolation: models.Attr[
        types.Auto | Literal["sRGB", "linearRGB"] | types.Inherit
    ] = None


class ColorProfile(common.Attr):
    color_profile: models.Attr[
        types.Auto
        | Literal["sRGB"]
        | types.Name
        | types.Iri
        | types.Inherit
    ] = None


class ColorRendering(common.Attr):
    color_rendering: models.Attr[
        types.Auto
        | Literal["optimizeSpeed", "optimizeQuality"]
        | types.Inherit
    ] = None


class Color(common.Attr):
    color: models.Attr[types.Color | types.Inherit] = None


class Cursor(common.Attr):
    cursor: models.Attr[
        types.CursorValue | types.Inherit | models.List[types.CursorValue]
    ] = None


class Direction(common.Attr):
    direction: models.Attr[Literal["ltr", "rtl"] | types.Inherit] = None


class Display(common.Attr):
    display: models.Attr[
        Literal[
            "block",
            "compact",
            "inline-table",
            "inline",
            "list-item",
            "marker",
            "run-in",
            "table-caption",
            "table-cell",
            "table-column-group",
            "table-column",
            "table-footer-group",
            "table-header-group",
            "table-row-group",
            "table-row",
            "table",
        ]
        | types.None_
        | types.Inherit
    ] = None


class DominantBaseline(common.Attr):
    dominant_baseline: models.Attr[
        types.Auto
        | Literal[
            "use-script",
            "no-change",
            "reset-size",
            "ideographic",
            "alphabetic",
            "hanging",
            "mathematical",
            "central",
            "middle",
            "text-after-edge",
            "text-before-edge",
        ]
        | types.Inherit
    ] = None


class EnableBackground(common.Attr):
    enable_background: models.Attr[
        Literal["accumulate"]
        | types.Inherit
        | types.Unparsed  # new [ <x> <y> <width> <height> ]
    ] = None


class FillOpacity(common.Attr):
    fill_opacity: models.Attr[types.OpacityValue | types.Inherit] = None


class FillRule(common.Attr):
    fill_rule: models.Attr[
        Literal["nonzero", "evenodd"] | types.Inherit
    ] = None


class Filter(common.Attr):
    filter: models.Attr[types.FuncIri | types.None_ | types.Inherit] = None


class FloodColor(common.Attr):
    flood_color: models.Attr[
        Literal["currentColor"]
        | types.Color
        | types.Inherit
        | types.Unparsed  # <color> [<icccolor>]
    ] = None


class FloodOpacity(common.Attr):
    flood_opacity: models.Attr[types.OpacityValue | types.Inherit] = None


class FontSizeAdjust(common.Attr):
    font_size_adjust: models.Attr[
        types.Number | types.None_ | types.Inherit
    ] = None


class FontSize(common.Attr):
    font_size: models.Attr[
        types.AbsoluteSize
        | types.RelativeSize
        | types.Length
        | types.Percentage
        | types.Inherit
    ] = None


class GlyphOrientationHorizontal(common.Attr):
    glyph_orientation_horizontal: models.Attr[
        types.Angle | types.Inherit
    ] = None


class GlyphOrientationVertical(common.Attr):
    glyph_orientation_vertical: models.Attr[
        types.Auto | types.Angle | types.Inherit
    ] = None


class ImageRendering(common.Attr):
    image_rendering: models.Attr[
        types.Auto
        | Literal["optimizeSpeed", "optimizeQuality"]
        | types.Inherit
    ] = None


class Kerning(common.Attr):
    kerning: models.Attr[types.Auto | types.Length | types.Inherit] = None


class LetterSpacing(common.Attr):
    letter_spacing: models.Attr[
        Literal["normal"] | types.Length | types.Inherit
    ] = None


class LightingColor(common.Attr):
    lighting_color: models.Attr[
        Literal["currentColor"]
        | types.Color
        | types.Inherit
        | types.Unparsed
    ] = None


class MarkerEnd(common.Attr):
    marker_end: models.Attr[
        types.FuncIri | types.None_ | types.Inherit
    ] = None


class MarkerMid(common.Attr):
    marker_mid: models.Attr[
        types.FuncIri | types.None_ | types.Inherit
    ] = None


class MarkerStart(common.Attr):
    marker_start: models.Attr[
        types.FuncIri | types.None_ | types.Inherit
    ] = None


class Mask(common.Attr):
    mask: models.Attr[types.FuncIri | types.None_ | types.Inherit] = None


class Opacity(common.Attr):
    opacity: models.Attr[types.OpacityValue | types.Inherit] = None


class Overflow(common.Attr):
    overflow: models.Attr[
        Literal["visible", "hidden", "scroll"] | types.Auto | types.Inherit
    ] = None


class PointerEvents(common.Attr):
    pointer_events: models.Attr[
        Literal[
            "visiblePainted",
            "visibleFill",
            "visibleStroke",
            "visible",
            "painted",
            "fill",
            "stroke",
        ]
        | types.All
        | types.None_
        | types.Inherit
    ] = None


class ShapeRendering(common.Attr):
    shape_rendering: models.Attr[
        types.Auto
        | Literal["optimizeSpeed", "crispEdges", "geometricPrecision"]
        | types.Inherit
    ] = None


class StopColor(common.Attr):
    stop_color: models.Attr[
        Literal["currentColor"]
        | types.Inherit
        | types.Unparsed  # <color> <icccolor>
    ] = None


class StopOpacity(common.Attr):
    stop_opacity: models.Attr[types.OpacityValue | types.Inherit] = None


class StrokeDasharray(common.Attr):
    stroke_dasharray: models.Attr[
        types.None_ | types.Dasharray | types.Inherit
    ] = None


class StrokeDashoffset(common.Attr):
    stroke_dashoffset: models.Attr[
        types.Percentage | types.Length | types.Inherit
    ] = None


class StrokeLinecap(common.Attr):
    stroke_linecap: models.Attr[
        Literal["butt", "round", "square"] | types.Inherit
    ] = None


class StrokeLinejoin(common.Attr):
    stroke_linejoin: models.Attr[
        Literal["miter", "round", "bevel"] | types.Inherit
    ] = None


class StrokeMiterlimit(common.Attr):
    stroke_miterlimit: models.Attr[types.Miterlimit | types.Inherit] = None


class StrokeOpacity(common.Attr):
    stroke_opacity: models.Attr[types.OpacityValue | types.Inherit] = None


class StrokeWidth(common.Attr):
    stroke_width: models.Attr[
        types.Percentage | types.Length | types.Inherit
    ] = None


class Stroke(common.Attr):
    stroke: models.Attr[types.Paint] = None


class TextAnchor(common.Attr):
    text_anchor: models.Attr[
        Literal["start", "middle", "end"] | types.Inherit
    ] = None


class TextDecoration(common.Attr):
    text_decoration: models.Attr[
        types.None_
        | Literal["underline", "overline", "line-through", "blink"]
        | types.Inherit
    ] = None


class TextRendering(common.Attr):
    text_rendering: models.Attr[
        types.Auto
        | Literal[
            "optimizeSpeed", "optimizeLegibility", "geometricPrecision"
        ]
        | types.Inherit
    ] = None


class UnicodeBidi(common.Attr):
    unicode_bidi: models.Attr[
        Literal["normal", "embed", "bidi-override"] | types.Inherit
    ] = None


class Visibility(common.Attr):
    visibility: models.Attr[
        Literal["visible", "hidden", "collapse"] | types.Inherit
    ] = None


class WordSpacing(common.Attr):
    word_spacing: models.Attr[
        Literal["normal"] | types.Length | types.Inherit
    ] = None


class WritingMode(common.Attr):
    writing_mode: models.Attr[
        Literal["lr-tb", "rl-tb", "tb-rl", "lr", "rl", "tb"]
        | types.Inherit
    ] = None



from svglab.elements.common import (
    Element,
    PairedTag,
    Tag,
    TextElement,
    tag_name,
)
from svglab.elements.names import (
    TAG_NAME_TO_NORMALIZED,
    TAG_NAMES,
    TagName,
)
from svglab.elements.tags import (
    A,
    AltGlyph,
    AltGlyphDef,
    AltGlyphItem,
    Animate,
    AnimateColor,
    AnimateMotion,
    AnimateTransform,
    Circle,
    ClipPath,
    ColorProfile,
    Cursor,
    Defs,
    Desc,
    Ellipse,
    FeBlend,
    FeColorMatrix,
    FeComponentTransfer,
    FeComposite,
    FeConvolveMatrix,
    FeDiffuseLighting,
    FeDisplacementMap,
    FeDistantLight,
    FeFlood,
    FeFuncA,
    FeFuncB,
    FeFuncG,
    FeFuncR,
    FeGaussianBlur,
    FeImage,
    FeMerge,
    FeMergeNode,
    FeMorphology,
    FeOffset,
    FePointLight,
    FeSpecularLighting,
    FeSpotLight,
    FeTile,
    FeTurbulence,
    Filter,
    Font,
    FontFace,
    FontFaceFormat,
    FontFaceName,
    FontFaceSrc,
    FontFaceUri,
    ForeignObject,
    G,
    Glyph,
    GlyphRef,
    Hkern,
    Image,
    Line,
    LinearGradient,
    Marker,
    Mask,
    Metadata,
    MissingGlyph,
    Mpath,
    Path,
    Pattern,
    Polygon,
    Polyline,
    RadialGradient,
    Rect,
    Script,
    Set,
    Stop,
    Style,
    Svg,
    Switch,
    Symbol,
    Text,
    TextPath,
    Title,
    Tref,
    Tspan,
    Use,
    View,
    Vkern,
)
from svglab.elements.text_elements import CData, Comment, RawText


__all__ = [
    "TAG_NAMES",
    "TAG_NAME_TO_NORMALIZED",
    "A",
    "AltGlyph",
    "AltGlyphDef",
    "AltGlyphItem",
    "Animate",
    "AnimateColor",
    "AnimateMotion",
    "AnimateTransform",
    "CData",
    "Circle",
    "ClipPath",
    "ColorProfile",
    "Comment",
    "Cursor",
    "Defs",
    "Desc",
    "Element",
    "Ellipse",
    "FeBlend",
    "FeColorMatrix",
    "FeComponentTransfer",
    "FeComposite",
    "FeConvolveMatrix",
    "FeDiffuseLighting",
    "FeDisplacementMap",
    "FeDistantLight",
    "FeFlood",
    "FeFuncA",
    "FeFuncB",
    "FeFuncG",
    "FeFuncR",
    "FeGaussianBlur",
    "FeImage",
    "FeMerge",
    "FeMergeNode",
    "FeMorphology",
    "FeOffset",
    "FePointLight",
    "FeSpecularLighting",
    "FeSpotLight",
    "FeTile",
    "FeTurbulence",
    "Filter",
    "Font",
    "FontFace",
    "FontFaceFormat",
    "FontFaceName",
    "FontFaceSrc",
    "FontFaceUri",
    "ForeignObject",
    "G",
    "G",
    "Glyph",
    "GlyphRef",
    "Hkern",
    "Image",
    "Line",
    "LinearGradient",
    "Marker",
    "Mask",
    "Metadata",
    "MissingGlyph",
    "Mpath",
    "PairedTag",
    "Path",
    "Pattern",
    "Polygon",
    "Polyline",
    "RadialGradient",
    "RawText",
    "RawText",
    "Rect",
    "Rect",
    "Script",
    "Set",
    "Stop",
    "Style",
    "Svg",
    "Svg",
    "Switch",
    "Symbol",
    "Tag",
    "TagName",
    "Text",
    "TextElement",
    "TextPath",
    "Title",
    "Tref",
    "Tspan",
    "Use",
    "View",
    "Vkern",
    "tag_name",
]



import contextlib
import os
import pathlib
from typing import final, overload

from svglab import serialize, utils
from svglab.attrs import groups, regular
from svglab.elements import traits


# ! WARNING: `Element` and `ContainerElement` must always go last


@final
class A(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    regular.Target,
    regular.Transform,
    traits.ContainerElement,
):
    pass


@final
class AltGlyph(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.DxListOfLengths,
    regular.DyListOfLengths,
    regular.ExternalResourcesRequired,
    regular.Format,
    regular.GlyphRef,
    regular.RotateListOfNumbers,
    regular.Style,
    regular.XListOfCoordinates,
    regular.YListOfCoordinates,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class AltGlyphDef(traits.Element):
    pass


@final
class AltGlyphItem(traits.Element):
    pass


@final
class Animate(
    groups.AnimationAddition,
    groups.AnimationAttributeTarget,
    groups.AnimationValue,
    groups.ConditionalProcessing,
    groups.Xlink,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateColor(
    groups.AnimationAddition,
    groups.AnimationAttributeTarget,
    groups.AnimationValue,
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateMotion(
    groups.AnimationAddition,
    groups.AnimationValue,
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.KeyPoints,
    regular.Origin,
    regular.Path,
    regular.RotateNumberAutoAutoReverse,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateTransform(
    groups.AnimationAddition,
    groups.AnimationAttributeTarget,
    groups.AnimationValue,
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.TypeAnimateTransform,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class Circle(
    groups.ConditionalProcessing,
    regular.Class,
    regular.Cx,
    regular.Cy,
    regular.ExternalResourcesRequired,
    regular.R,
    regular.Style,
    regular.Transform,
    traits.BasicShape,
    traits.Element,
):
    pass


@final
class ClipPath(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ClipPathUnits,
    regular.ExternalResourcesRequired,
    regular.Style,
    regular.Transform,
    traits.Element,
):
    pass


@final
class ColorProfile(
    groups.Xlink,
    regular.Local,
    regular.NameName,
    regular.RenderingIntent,
    traits.Element,
):
    pass


@final
class Cursor(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.Element,
):
    pass


@final
class Defs(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    regular.Transform,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Desc(
    regular.Class, regular.Style, traits.DescriptiveElement, traits.Element
):
    pass


@final
class Ellipse(
    groups.ConditionalProcessing,
    regular.Class,
    regular.Cx,
    regular.Cy,
    regular.ExternalResourcesRequired,
    regular.Rx,
    regular.Ry,
    regular.Style,
    regular.Transform,
    traits.BasicShape,
    traits.Element,
):
    pass


@final
class FeBlend(
    regular.Class,
    regular.In,
    regular.In2,
    regular.Mode,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeColorMatrix(
    regular.Class,
    regular.In,
    regular.In2,
    regular.Style,
    regular.TypeFeColorMatrix,
    regular.ValuesListOfNumbers,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeComponentTransfer(
    regular.Class,
    regular.In,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeComposite(
    regular.Class,
    regular.In,
    regular.In2,
    regular.K1,
    regular.K2,
    regular.K3,
    regular.K4,
    regular.OperatorFeComposite,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeConvolveMatrix(
    regular.Bias,
    regular.Class,
    regular.Divisor,
    regular.In,
    regular.KernelMatrix,
    regular.KernelUnitLength,
    regular.Order,
    regular.PreserveAlpha,
    regular.Style,
    regular.TargetX,
    regular.TargetY,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeDiffuseLighting(
    regular.Class,
    regular.DiffuseConstant,
    regular.In,
    regular.KernelUnitLength,
    regular.Style,
    regular.SurfaceScale,
    traits.Element,
):
    pass


@final
class FeDisplacementMap(
    regular.Class,
    regular.In,
    regular.In2,
    regular.Scale,
    regular.Style,
    regular.XChannelSelector,
    regular.YChannelSelector,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeDistantLight(
    regular.Azimuth,
    regular.Elevation,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeFlood(
    regular.Class,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeFuncA(groups.TransferFunction, traits.Element):
    pass


@final
class FeFuncB(groups.TransferFunction, traits.Element):
    pass


@final
class FeFuncG(groups.TransferFunction, traits.Element):
    pass


@final
class FeFuncR(groups.TransferFunction, traits.Element):
    pass


@final
class FeGaussianBlur(
    regular.Class,
    regular.In,
    regular.StdDeviation,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeImage(
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.PreserveAspectRatio,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeMerge(
    regular.Class,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeMergeNode(regular.In, traits.Element):
    pass


@final
class FeMorphology(
    regular.Class,
    regular.In,
    regular.OperatorFeMorphology,
    regular.Radius,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeOffset(
    regular.Class,
    regular.DxNumber,
    regular.DyNumber,
    regular.In,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FePointLight(
    regular.XNumber,
    regular.YNumber,
    regular.Z,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeSpecularLighting(
    regular.Class,
    regular.In,
    regular.KernelUnitLength,
    regular.SpecularConstant,
    regular.SpecularExponent,
    regular.Style,
    regular.SurfaceScale,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeSpotLight(
    regular.LimitingConeAngle,
    regular.PointsAtX,
    regular.PointsAtY,
    regular.PointsAtZ,
    regular.SpecularExponent,
    regular.XNumber,
    regular.YNumber,
    regular.Z,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeTile(
    regular.Class,
    regular.In,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeTurbulence(
    regular.BaseFrequency,
    regular.Class,
    regular.NumOctaves,
    regular.Seed,
    regular.StitchTiles,
    regular.Style,
    regular.TypeFeTurbluence,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class Filter(
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.FilterRes,
    regular.FilterUnits,
    regular.Height,
    regular.PrimitiveUnits,
    regular.Style,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.Element,
):
    pass


@final
class Font(
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.HorizAdvX,
    regular.HorizOriginX,
    regular.HorizOriginY,
    regular.Style,
    regular.VertAdvY,
    regular.VertOriginX,
    regular.VertOriginY,
    traits.Element,
):
    pass


@final
class FontFace(
    regular.AccentHeight,
    regular.Alphabetic,
    regular.Ascent,
    regular.Bbox,
    regular.CapHeight,
    regular.Descent,
    regular.FontFamily,
    regular.FontSize,
    regular.Hanging,
    regular.Ideographic,
    regular.Mathematical,
    regular.OverlinePosition,
    regular.OverlineThickness,
    regular.Panose1,
    regular.Slope,
    regular.Stemh,
    regular.Stemv,
    regular.StrikethroughPosition,
    regular.StrikethroughThickness,
    regular.UnderlinePosition,
    regular.UnderlineThickness,
    regular.UnicodeRange,
    regular.UnitsPerEm,
    regular.VAlphabetic,
    regular.VHanging,
    regular.VIdeographic,
    regular.VMathematical,
    regular.Widths,
    regular.XHeight,
    traits.Element,
):
    pass


@final
class FontFaceFormat(regular.String, traits.Element):
    pass


@final
class FontFaceName(regular.String, traits.Element):
    pass


@final
class FontFaceSrc(traits.Element):
    pass


@final
class FontFaceUri(groups.Xlink, traits.Element):
    pass


@final
class ForeignObject(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.Style,
    regular.Transform,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.Element,
):
    pass


@final
class G(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    regular.Transform,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Glyph(
    regular.ArabicForm,
    regular.Class,
    regular.D,
    regular.GlyphName,
    regular.HorizAdvX,
    regular.Lang,
    regular.Orientation,
    regular.Style,
    regular.Unicode,
    regular.VertAdvY,
    regular.VertOriginX,
    regular.VertOriginY,
    traits.ContainerElement,
):
    pass


@final
class GlyphRef(
    groups.Xlink,
    regular.Class,
    regular.DxNumber,
    regular.DyNumber,
    regular.Format,
    regular.GlyphRef,
    regular.Style,
    regular.XNumber,
    regular.YNumber,
    traits.Element,
):
    pass


@final
class Hkern(
    regular.G1,
    regular.G2,
    regular.K,
    regular.U1,
    regular.U2,
    traits.Element,
):
    pass


@final
class Image(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.PreserveAspectRatio,
    regular.Style,
    regular.Transform,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.GraphicsElement,
    traits.GraphicsReferencingElement,
    traits.Element,
):
    pass


@final
class Line(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    regular.Transform,
    regular.X1,
    regular.X2,
    regular.Y1,
    regular.Y2,
    traits.BasicShape,
    traits.Element,
):
    pass


@final
class LinearGradient(
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.GradientTransform,
    regular.GradientUnits,
    regular.SpreadMethod,
    regular.Style,
    regular.X1,
    regular.X2,
    regular.Y1,
    regular.Y2,
    traits.GradientElement,
    traits.Element,
):
    pass


@final
class Marker(
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.MarkerHeight,
    regular.MarkerUnits,
    regular.MarkerWidth,
    regular.Orient,
    regular.PreserveAspectRatio,
    regular.RefX,
    regular.RefY,
    regular.Style,
    regular.ViewBox,
    traits.ContainerElement,
):
    pass


@final
class Mask(
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.MaskContentUnits,
    regular.MaskUnits,
    regular.Style,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.ContainerElement,
):
    pass


@final
class Metadata(traits.DescriptiveElement, traits.Element):
    pass


@final
class MissingGlyph(
    regular.Class,
    regular.D,
    regular.HorizAdvX,
    regular.Style,
    regular.Transform,
    regular.VertAdvY,
    regular.VertOriginX,
    regular.VertOriginY,
    traits.ContainerElement,
):
    pass


@final
class Mpath(
    groups.Xlink, regular.ExternalResourcesRequired, traits.Element
):
    pass


@final
class Path(
    groups.ConditionalProcessing,
    regular.Class,
    regular.D,
    regular.ExternalResourcesRequired,
    regular.PathLength,
    regular.Style,
    regular.Transform,
    traits.Shape,
    traits.Element,
):
    pass


@final
class Pattern(
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.PatternContentUnits,
    regular.PatternTransform,
    regular.PatternUnits,
    regular.PreserveAspectRatio,
    regular.Style,
    regular.ViewBox,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.ContainerElement,
):
    pass


@final
class Polygon(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Points,
    regular.Style,
    regular.Transform,
    traits.BasicShape,
    traits.Element,
):
    pass


@final
class Polyline(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Points,
    regular.Style,
    regular.Transform,
    traits.BasicShape,
    traits.Element,
):
    pass


@final
class RadialGradient(
    groups.Xlink,
    regular.Class,
    regular.Cx,
    regular.Cy,
    regular.ExternalResourcesRequired,
    regular.Fx,
    regular.Fy,
    regular.GradientTransform,
    regular.GradientUnits,
    regular.R,
    regular.SpreadMethod,
    regular.Style,
    traits.GradientElement,
    traits.Element,
):
    pass


@final
class Rect(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.Rx,
    regular.Ry,
    regular.Style,
    regular.Transform,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.BasicShape,
    traits.Element,
):
    pass


@final
class Script(
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.TypeContentType,
    traits.Element,
):
    pass


@final
class Set(
    groups.AnimationAttributeTarget,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.To,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class Stop(
    regular.Class,
    regular.OffsetNumberPercentage,
    regular.Style,
    traits.Element,
):
    pass


@final
class Style(
    regular.Media, regular.Title, regular.TypeContentType, traits.Element
):
    pass


@final
class Svg(
    groups.ConditionalProcessing,
    groups.DocumentEvents,
    regular.BaseProfile,
    regular.Class,
    regular.ContentScriptType,
    regular.ContentStyleType,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.PreserveAspectRatio,
    regular.Style,
    regular.Version,
    regular.ViewBox,
    regular.Width,
    regular.XCoordinate,
    regular.Xmlns,
    regular.YCoordinate,
    regular.ZoomAndPan,
    traits.StructuralElement,
    traits.ContainerElement,
):
    @overload
    def save(
        self,
        path: str | os.PathLike[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None: ...

    @overload
    def save(
        self,
        file: utils.SupportsWrite[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None: ...

    def save(
        self,
        path_or_file: str | os.PathLike[str] | utils.SupportsWrite[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None:
        """Convert the SVG document fragment to XML and write it to a file.

        Args:
        path_or_file: The path to the file to save the XML to,
        or a file-like object.
        pretty: Whether to produce pretty-printed XML.
        indent: The number of spaces to indent each level of the document.
        trailing_newline: Whether to add a trailing newline to the file.
        formatter: The formatter to use for serialization.

        Examples:
        >>> import sys
        >>> svg = Svg(id="foo").add_child(Rect())
        >>> formatter = serialize.Formatter(indent=4)
        >>> svg.save(
        ...     sys.stdout,
        ...     pretty=True,
        ...     trailing_newline=False,
        ...     formatter=formatter,
        ... )
        <svg id="foo" xmlns="http://www.w3.org/2000/svg">
            <rect/>
        </svg>

        """
        with contextlib.ExitStack() as stack:
            output = self.to_xml(pretty=pretty, formatter=formatter)
            file: utils.SupportsWrite[str]

            match path_or_file:
                case str() | os.PathLike() as path:
                    file = stack.enter_context(
                        pathlib.Path(path).open("w")
                    )
                case utils.SupportsWrite() as file:
                    pass

            file.write(output)

            if trailing_newline:
                file.write("\n")


@final
class Switch(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    regular.Transform,
    traits.ContainerElement,
):
    pass


@final
class Symbol(
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.PreserveAspectRatio,
    regular.Style,
    regular.ViewBox,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Text(
    groups.ConditionalProcessing,
    regular.Class,
    regular.DxListOfLengths,
    regular.DyListOfLengths,
    regular.ExternalResourcesRequired,
    regular.LengthAdjust,
    regular.RotateListOfNumbers,
    regular.Style,
    regular.TextLength,
    regular.Transform,
    regular.XListOfCoordinates,
    regular.YListOfCoordinates,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class TextPath(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Method,
    regular.Spacing,
    regular.StartOffset,
    regular.Style,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Title(
    regular.Class, regular.Style, traits.DescriptiveElement, traits.Element
):
    pass


@final
class Tref(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Tspan(
    groups.ConditionalProcessing,
    regular.Class,
    regular.DxListOfLengths,
    regular.DyListOfLengths,
    regular.ExternalResourcesRequired,
    regular.LengthAdjust,
    regular.RotateListOfNumbers,
    regular.Style,
    regular.TextLength,
    regular.XListOfCoordinates,
    regular.YListOfCoordinates,
    traits.TextContentBlockElement,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Use(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.Style,
    regular.Transform,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.GraphicsElement,
    traits.GraphicsReferencingElement,
    traits.StructuralElement,
    traits.Element,
):
    pass


@final
class View(
    regular.ExternalResourcesRequired,
    regular.PreserveAspectRatio,
    regular.ViewBox,
    regular.ViewTarget,
    regular.ZoomAndPan,
    traits.Element,
):
    pass


@final
class Vkern(
    regular.G1,
    regular.G2,
    regular.K,
    regular.U1,
    regular.U2,
    traits.Element,
):
    pass



#!/usr/bin/env python3

# ruff: noqa: T201

import sys

from svglab import (
    CData,
    Comment,
    G,
    Path,
    Polyline,
    RawText,
    Rect,
    parse_svg,
)
from svglab.attrparse import Color, D, Length, Point, SkewX, Translate
from svglab.serialize import Formatter, set_formatter


set_formatter(Formatter(indent=4, max_precision=2, color_mode="rgb"))


def main() -> None:
    # Parse an existing SVG file
    svg = parse_svg(
        """
        <svg xmlns="http://www.w3.org/2000/svg">
          <g>
              <rect
                id="background"
                width="100cm"
                height="100%"
                transform="rotate(45)"
                stroke="red"
              />
              <rect color="hsl(0, 100%, 100%)"/>
              <!-- This is a comment -->
              <![CDATA[.background { fill: blue; }]]>
              Hello SVG!
              <path d="M 10,10 L 100,100 Q 100,100 50,50 Z"/>
              <polygon points="0,0 100,0 100,100 0,100"/>
          </g>
        </svg>
    """
    )

    print(svg)

    # Create an element programmatically
    group = G().add_children(
        Rect(
            width=Length(15, "px"),
            height=Length(20),
            transform=[SkewX(45.123), Translate(10, 20)],
            color=Color("#ff0000"),
        ),
        Comment("This is a comment"),
        CData(".background { fill: blue; }"),
        RawText("Hello SVG!"),
        Path(
            d=D()
            .move_to(Point(10, 10))
            .line_to(Point(100, 100))
            .quadratic_bezier_to(Point(100, 100), Point(50, 50))
            .move_to(Point(50, 50))
            .cubic_bezier_to(
                Point(100, 100), Point(100, 100), Point(10, 10)
            )
            .arc_to(
                Point(50, 50), 90, Point(100, 100), large=True, sweep=False
            )
            .close()
        ),
        Polyline(
            points=[
                Point(0, 0),
                Point(100, 0),
                Point(100, 100),
                Point(0, 100),
            ],
            stroke_linecap="square",
            opacity=0.5,
        ),
    )

    # Add the element to the SVG
    svg.add_child(group)

    # Manipulate attributes
    print(svg.xmlns)  # http://www.w3.org/2000/svg
    svg.x = Length(10, "px")

    # Save to a file
    svg.save(sys.stdout)

    print(svg.find(G).find(Rect).width)
    print(*svg.find_all(Rect), sep="\n")


if __name__ == "__main__":
    main()



import collections
from typing import Final, Literal, TypeAlias, cast

import bs4

from svglab import elements, utils


Parser: TypeAlias = Literal["html.parser", "lxml", "lxml-xml", "html5lib"]
""" Type for parsers supported by BeautifulSoup. """


DEFAULT_PARSER: Final[Parser] = "lxml-xml"


_TAG_NAME_TO_CLASS: Final = {
    elements.tag_name(cls): cls
    for cls in utils.get_all_subclasses(elements.Tag)
    if cls.__name__ in elements.TAG_NAME_TO_NORMALIZED.inverse
}

_BS_TO_TEXT_ELEMENT: Final[
    dict[
        type[bs4.NavigableString],
        type[elements.CData | elements.Comment | elements.RawText],
    ]
] = {
    bs4.CData: elements.CData,
    bs4.Comment: elements.Comment,
    bs4.NavigableString: elements.RawText,
}


def _get_root_svg_fragments(soup: bs4.Tag) -> list[bs4.Tag]:
    """Find all root SVG fragments in the given BeautifulSoup object.

    The function performs a breadth-first search until it finds an
    SVG fragment. It then returns a list of all SVG fragments found in the same
    depth of the tree. This allows us to consider an SVG fragment as a root
    element when using HTML parsers that implicitly wrap the document in
    certain HTML tags (e.g., <html>).

    Args:
        soup: A BeautifulSoup tag object representing the root of the document.

    Returns:
        A list of SVG fragments found in the document.

    Examples:
        >>> soup = bs4.BeautifulSoup("<svg><rect/></svg>", features="lxml-xml")
        >>> _get_root_svg_fragments(soup)
        [<svg><rect/></svg>]
        >>> soup = bs4.BeautifulSoup(
        ...     "<svg><rect/></svg>", features="html.parser"
        ... )
        >>> _get_root_svg_fragments(soup)
        [<svg><rect></rect></svg>]

    """
    queue: collections.deque[bs4.Tag] = collections.deque([soup])

    while queue:
        node = queue.popleft()

        svg_fragments = node.find_all("svg")

        if svg_fragments:
            return svg_fragments

        queue.extend(
            child for child in node.children if isinstance(child, bs4.Tag)
        )

    return []


def _convert_element(backend: bs4.PageElement) -> elements.Element | None:
    """Convert a BeautifulSoup element to an `Element` instance.

    Args:
        backend: A BeautifulSoup element to convert.

    Returns:
        An `Element` instance representing the given BeautifulSoup element.

    Raises:
        TypeError: If the given element cannot be converted.

    """
    match backend:
        case bs4.NavigableString():
            cls = _BS_TO_TEXT_ELEMENT.get(type(backend))

            if cls is None:
                return None

            text = backend.get_text(strip=True)

            if not text:
                return None

            return cls(text)
        case bs4.Tag():
            tag_class = _TAG_NAME_TO_CLASS[
                cast(elements.TagName, backend.name)
            ]

            tag = tag_class.model_validate(
                {"prefix": backend.prefix, **backend.attrs}, strict=False
            )

            if isinstance(tag, elements.PairedTag):
                for child in backend.children:
                    grandchild = _convert_element(child)

                    if grandchild is not None:
                        tag.add_child(grandchild)
            elif not utils.is_empty(backend.contents):
                msg = f"Unpaired tag {type(tag)} cannot have children."
                raise TypeError(msg)

            return tag
        case _:
            return None


def parse_svg(
    markup: str
    | bytes
    | utils.SupportsRead[str]
    | utils.SupportsRead[bytes],
    /,
    *,
    parser: Parser = DEFAULT_PARSER,
) -> elements.Svg:
    """Parse an SVG document.

    The document must be a valid XML document containing a single SVG
    document fragment.

    Args:
        markup: A string or a file-like object representing markup
        to be parsed.
        parser: The name of the parser to use. Defaults to 'lxml-xml'.

    Returns:
        The parsed SVG document in the form of an `Svg` instance.

    Raises:
        ValueError: If the markup does not contain a single SVG
        document fragment

    Examples:
        >>> svg = parse_svg("<svg><rect/></svg>")
        >>> type(svg).__name__
        'Svg'

    """
    soup = bs4.BeautifulSoup(markup, features=parser)

    svg_fragments = _get_root_svg_fragments(soup)

    if len(svg_fragments) != 1:
        msg = (
            f"Expected 1 <svg> element, found {len(svg_fragments)}."
            " This does not look like a valid SVG."
        )

        raise ValueError(msg)

    svg = _convert_element(svg_fragments[0])

    if not isinstance(svg, elements.Svg):
        msg = f"Expected an <svg> element, found {type(svg).__name__}."
        raise TypeError(msg)

    return svg



import abc
import functools
import re
import reprlib
from collections.abc import Callable
from typing import (
    Annotated,
    Protocol,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)

import pydantic
import pydantic_core.core_schema
from pydantic import Field
from typing_extensions import Self, override


_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_T_seq = TypeVar("_T_seq", list[str], tuple[str])

_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")

KwOnly: TypeAlias = Annotated[_T_co, Field(kw_only=True)]
""" Pydantic field for a keyword-only attribute. """

Attr: TypeAlias = KwOnly[_T_co | None]
""" Pydantic field for an attribute. """


def _parse_list(text: str, /, collection: type[_T_seq] = list) -> _T_seq:
    """Parse a string into a list of strings.

    Items are separated by whitespace or commas.

    Args:
        text: The string to parse.
        collection: The type of collection to return.

    Returns:
        A collection of strings.

    Examples:
        >>> _parse_list("a b c")
        ['a', 'b', 'c']
        >>> _parse_list("a, b, c")
        ['a', 'b', 'c']
        >>> _parse_list("a b, c")
        ['a', 'b', 'c']
        >>> _parse_list("a,b,c")
        ['a', 'b', 'c']
        >>> _parse_list("")
        []

    """
    result = (part for part in re.split(r"\s+|\s*,\s*", text) if part)
    return collection(result)


def get_validator(
    func: Callable[[str], object], /
) -> pydantic.BeforeValidator:
    def validator(value: object) -> object:
        if isinstance(value, str):
            return func(value)

        return value

    return pydantic.BeforeValidator(validator)


List: TypeAlias = Annotated[
    list[_T],
    get_validator(functools.partial(_parse_list, collection=list)),
]
"""Pydantic field for a list of strings. Uses `_parse_list` as a validator."""

Tuple: TypeAlias = Annotated[
    _T, get_validator(functools.partial(_parse_list, collection=tuple))
]
"""Pydantic field for a tuple of strings. Uses `_parse_list` as a validator."""

# unfortunately, there doesn't seem to be a better way to do this
# see https://github.com/python/typing/issues/779
Tuple1: TypeAlias = Tuple[tuple[_T1]]
Tuple2: TypeAlias = Tuple[tuple[_T1, _T2]]
Tuple3: TypeAlias = Tuple[tuple[_T1, _T2, _T3]]
Tuple4: TypeAlias = Tuple[tuple[_T1, _T2, _T3, _T4]]


class BaseModel(pydantic.BaseModel):
    """Pydantic BaseModel with sane defaults and a few extra tweaks."""

    model_config = pydantic.ConfigDict(
        allow_inf_nan=False,
        extra="forbid",
        strict=True,
        validate_assignment=True,
        validate_default=True,
        validate_return=True,
    )

    @override
    def __str__(self) -> str:
        return repr(self)

    # patch pydantic's __repr__ so that is doesn't break on cyclic models
    # see https://github.com/pydantic/pydantic/issues/9424
    @reprlib.recursive_repr()
    @override
    def __repr__(self) -> str:
        return super().__repr__()


@runtime_checkable
class PydanticCompatible(Protocol):
    """A protocol for classes that can be used as pydantic models."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: type, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.core_schema.CoreSchema: ...


class CustomModel(PydanticCompatible, metaclass=abc.ABCMeta):
    """A mixin for easy creation of custom pydantic-compatible classes.

    This class is a mixin for classes that need to be pydantic-compatible,
    but are not, for technical reasons, pydantic dataclasses or subclasses
    of `BaseModel` (for example collections).

    By implementing the `_validate` class method, the class can be used
    in pydantic models. `__get_pydantic_core_schema__` is created
    automatically and should not be overridden.
    """

    @classmethod
    @abc.abstractmethod
    def _validate(
        cls, value: object, info: pydantic_core.core_schema.ValidationInfo
    ) -> Self:
        """Validate the value and return a new instance of the class."""

    @override
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: type, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.core_schema.CoreSchema:
        del source_type, handler

        return (
            pydantic_core.core_schema.with_info_plain_validator_function(
                function=cls._validate
            )
        )



from __future__ import annotations

import abc
import collections
import reprlib
import sys
from collections.abc import Generator, Mapping
from typing import Final, SupportsIndex, TypeVar, cast, overload

import bs4
import pydantic
from typing_extensions import Self, override

from svglab import errors, models, serialize, utils
from svglab.attrs import names as attr_names
from svglab.elements import names


_T = TypeVar("_T")
_T_tag = TypeVar("_T_tag", bound="Tag")

_EMPTY_PARAM: Final = object()
"""A sentinel value for an empty parameter."""


def tag_name(tag: Tag | type[Tag], /) -> names.TagName:
    """Get the SVG tag name of the given tag or tag class.

    Args:
    tag: The tag or tag class.

    Returns:
    The SVG tag name.

    Examples:
    >>> from svglab import Rect
    >>> tag_name(Rect)
    'rect'

    """
    tag_cls = tag if isinstance(tag, type) else type(tag)

    return names.TAG_NAME_TO_NORMALIZED.inverse[tag_cls.__name__]


def _match_tag(tag: Tag, /, *, search: type[Tag] | names.TagName) -> bool:
    """Check if a tag matches the given search criteria.

    Args:
    tag: The tag to check.
    search: The search criteria. Can be a tag name or a tag class.

    Returns:
    `True` if the tag matches the search criteria, `False` otherwise.

    Examples:
    >>> from svglab import Rect
    >>> rect = Rect()
    >>> _match_tag(rect, search="rect")
    True
    >>> _match_tag(rect, search=Rect)
    True
    >>> _match_tag(rect, search="circle")
    False

    """
    if isinstance(search, type):
        return isinstance(tag, search)

    return tag_name(tag) == search


class Element(models.BaseModel, metaclass=abc.ABCMeta):
    """The base class of the SVG element hierarchy."""

    parent: Element | None = pydantic.Field(default=None, init=False)

    def to_xml(
        self,
        *,
        pretty: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> str:
        """Convert the element to XML.

        Args:
        pretty: Whether to produce pretty-printed XML.
        indent: The number of spaces to indent each level of the document.
        formatter: The formatter to use for serialization.

        Returns:
        The XML representation of the element.

        Examples:
        >>> from svglab import Rect
        >>> from svglab.attrparse import Length
        >>> rect = Rect(id="foo", stroke_linecap="round")
        >>> rect.to_xml()
        '<rect id="foo" stroke-linecap="round"/>'

        """
        formatter = formatter or serialize.get_current_formatter()

        with formatter:
            soup = self.to_beautifulsoup_object()
            return utils.beautifulsoup_to_str(
                soup, pretty=pretty, indent=formatter.indent
            )

    @abc.abstractmethod
    def to_beautifulsoup_object(self) -> bs4.PageElement:
        """Convert the element to a corresponding `BeautifulSoup` object."""

    @override
    def __eq__(self, other: object) -> bool:
        if self is other:
            return True

        if not isinstance(other, type(self)):
            return False

        return self._eq(other)

    @abc.abstractmethod
    def _eq(self, other: Self) -> bool: ...


class TextElement(Element, metaclass=abc.ABCMeta):
    """The base class of text-based elements.

    Text-based elements are elements that are represented by a single string.

    Common examples of text-based elements in XML are:
    - `CDATA` sections (`CData`)
    - comments (`Comment`)
    - text (`Text`)
    - processing instructions (for example, `<?xml version="1.0"?>`)
    - Document Type Definitions (DTDs; for example, `<!DOCTYPE html>`)
    """

    content: str = pydantic.Field(frozen=True, min_length=1)

    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        return f"{name}({self.content!r})"

    @override
    def _eq(self, other: Self) -> bool:
        return self.content == other.content


class Tag(Element, metaclass=abc.ABCMeta):
    """A tag.

    A tag is an element that has a name and a set of attributes.

    Tags can be of two types:
    - paired tags, which have children
      (for example, `<g></g>`; see `PairedTag`)
    - unpaired tags, which do not have children (for example, `<rect />`)

    Tags can have standard attributes (for example, `id`, `class`, `color`) and
    non-standard user-defined attributes.

    Example: `<rect id="foo" class="bar" color="red" />`
    ```
    """

    model_config = pydantic.ConfigDict(
        extra="allow",
        alias_generator=pydantic.AliasGenerator(
            validation_alias=lambda name: pydantic.AliasChoices(
                name,
                attr_names.ATTR_NAME_TO_NORMALIZED.inverse.get(name, name),
            ),
            serialization_alias=(
                lambda name: attr_names.ATTR_NAME_TO_NORMALIZED.inverse.get(
                    name, name
                )
            ),
        ),
    )

    prefix: str | None = None

    @pydantic.model_validator(mode="after")
    def __validate_extra(self) -> Tag:  # pyright: ignore[reportUnusedFunction]
        # model_extra cannot be None because extra is set to "allow"
        assert self.model_extra is not None, "model_extra is None"

        for key, value in self.model_extra.items():
            if not isinstance(value, str):
                msg = (
                    f"Non-standard attribute {key!r} must be of type str,"
                    f" got {type(value)}"
                )
                raise TypeError(msg)

        return self

    def __getitem__(self, key: str) -> str:
        assert self.model_extra is not None, "model_extra is None"
        value: str = self.model_extra[key]
        return value

    def __setitem__(self, key: str, value: str) -> None:
        assert self.model_extra is not None, "model_extra is None"
        self.model_extra[key] = value

    def __delitem__(self, key: str) -> None:
        assert self.model_extra is not None, "model_extra is None"
        del self.model_extra[key]

    def extra_attrs(self) -> Mapping[str, str]:
        assert self.model_extra is not None, "model_extra is None"
        return self.model_extra

    def standard_attrs(self) -> Mapping[attr_names.AttributeName, object]:
        dump = self.model_dump(by_alias=True, exclude_none=True)

        return {
            attr: getattr(self, attr_names.ATTR_NAME_TO_NORMALIZED[attr])
            for key, _ in dump.items()
            if (attr := cast(attr_names.AttributeName, key))
            in attr_names.ATTR_NAME_TO_NORMALIZED
        }

    def all_attrs(self) -> Mapping[str, object]:
        standard = cast(Mapping[str, object], self.standard_attrs())
        extra = self.extra_attrs()

        return {**standard, **extra}

    @reprlib.recursive_repr()
    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        attrs = dict(self.all_attrs())

        if isinstance(self, PairedTag):
            attrs["children"] = list(self.children)

        attr_repr = ", ".join(
            f"{key}={value!r}" for key, value in attrs.items()
        )

        return f"{name}({attr_repr})"

    @override
    def to_beautifulsoup_object(self) -> bs4.Tag:
        tag = bs4.Tag(
            name=tag_name(self),
            can_be_empty_element=True,
            prefix=self.prefix,
            is_xml=True,
        )

        for key, value in self.all_attrs().items():
            tag[key] = serialize.serialize_attr(value)

        return tag

    @override
    def _eq(self, other: Self) -> bool:
        return (
            self.prefix == other.prefix
            and self.all_attrs() == other.all_attrs()
        )


class PairedTag(Tag, metaclass=abc.ABCMeta):
    """A paired tag.

    A paired tag is a tag that can have children.

    Example: `<g><rect /></g>`
    """

    __children: list[Element] = pydantic.PrivateAttr(default_factory=list)

    @pydantic.computed_field
    @property
    def children(self) -> Generator[Element, None, None]:
        yield from self.__children

    @pydantic.computed_field
    @property
    def descendants(self) -> Generator[Element, None, None]:
        queue = collections.deque(self.children)

        while queue:
            child = queue.popleft()
            yield child

            if isinstance(child, PairedTag):
                queue.extend(child.children)

    @pydantic.computed_field
    @property
    def parents(self) -> Generator[Element, None, None]:
        curr = self.parent

        while curr is not None:
            yield curr
            curr = curr.parent

    @pydantic.computed_field
    @property
    def next_siblings(self) -> Generator[Element, None, None]:
        if self.parent is None or not isinstance(self.parent, PairedTag):
            return

        should_yield = False

        for sibling in self.parent.children:
            if should_yield:
                yield sibling
            elif sibling is self:
                should_yield = True

    @pydantic.computed_field
    @property
    def prev_siblings(self) -> Generator[Element, None, None]:
        if self.parent is None or not isinstance(self.parent, PairedTag):
            return

        for sibling in self.parent.children:
            if sibling is self:
                return

            yield sibling

    @pydantic.computed_field
    @property
    def siblings(self) -> Generator[Element, None, None]:
        yield from self.prev_siblings
        yield from self.next_siblings

    def add_child(
        self, child: Element, /, *, index: int | None = None
    ) -> Self:
        if child is self:
            raise ValueError("Cannot add a tag as a child of itself.")

        if child.parent is not None:
            raise ValueError(
                "Cannot add a child that already has a parent."
            )

        if index is None:
            self.__children.append(child)
        else:
            self.__children.insert(index, child)

        child.parent = self

        return self

    def add_children(self, *children: Element) -> Self:
        for child in children:
            self.add_child(child)

        return self

    def remove_child(self, child: Element, /) -> Element:
        self.__children.remove(child)
        child.parent = None

        return child

    def pop_child(self, index: SupportsIndex = -1, /) -> Element:
        child = self.__children.pop(index)
        child.parent = None

        return child

    def clear_children(self) -> None:
        for child in self.__children:
            child.parent = None

        self.__children.clear()

    def get_child(self, index: SupportsIndex = -1, /) -> Element:
        return self.__children[index]

    def get_child_index(
        self,
        child: Element,
        /,
        start: SupportsIndex = 0,
        stop: SupportsIndex = sys.maxsize,
    ) -> int:
        return self.__children.index(child, start, stop)

    @override
    def to_beautifulsoup_object(self) -> bs4.Tag:
        tag = super().to_beautifulsoup_object()
        tag.can_be_empty_element = False

        for child in self.children:
            tag.append(child.to_beautifulsoup_object())

        return tag

    @overload
    def find_all(
        self, *tags: type[_T_tag], recursive: bool = True
    ) -> Generator[_T_tag, None, None]: ...

    @overload
    def find_all(
        self, *tags: type[Tag] | names.TagName, recursive: bool = True
    ) -> Generator[Tag, None, None]: ...

    def find_all(
        self, *tags: type[Tag] | names.TagName, recursive: bool = True
    ) -> Generator[Tag, None, None]:
        """Find all tags that match the given search criteria.

        Args:
        tags: The tags to search for. Can be tag names or tag classes.
        recursive: If `False`, only search the direct children of the tag,
        otherwise search all descendants.

        Returns:
        An iterator over all tags that match the search criteria.

        Examples:
        >>> from svglab import G, Rect
        >>> g = G().add_children(Rect(), G().add_child(Rect()))
        >>> list(g.find_all("rect"))
        [Rect(), Rect()]
        >>> list(g.find_all(G))
        [G(children=[Rect()])]
        >>> list(g.find_all(Rect, recursive=False))
        [Rect()]
        >>> list(g.find_all(G, "rect"))
        [Rect(), G(children=[Rect()]), Rect()]

        """
        for child in self.descendants if recursive else self.children:
            if isinstance(child, Tag) and any(
                _match_tag(child, search=tag) for tag in tags
            ):
                yield child

    @overload
    def find(
        self, *tags: type[_T_tag], recursive: bool = True
    ) -> _T_tag: ...

    @overload
    def find(
        self, *tags: type[Tag] | names.TagName, recursive: bool = True
    ) -> Tag: ...

    @overload
    def find(
        self,
        *tags: type[_T_tag],
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> _T_tag | _T: ...

    @overload
    def find(
        self,
        *tags: type[Tag] | names.TagName,
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> Tag | _T: ...

    def find(
        self,
        *tags: type[Tag] | names.TagName,
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> Tag | _T:
        """Find the first tag that matches the given search criteria.

        Args:
        tags: The tags to search for. Can be tag names or tag classes.
        recursive: If `False`, only search the direct children of the tag,
        otherwise search all descendants.
        default: The default value to return if no tag matches the search
        criteria.

        Returns:
        The first tag that matches the search criteria.

        Raises:
        SvgElementNotFoundError: If no tag matches the search criteria and no
        default value is provided.

        Examples:
        >>> from svglab import G, Rect
        >>> g = G().add_children(
        ...     Rect(id="foo"), G().add_child(Rect(id="bar"))
        ... )
        >>> g.find("rect")
        Rect(id='foo')
        >>> g.find(G)
        G(children=[Rect(id='bar')])
        >>> g.find("circle", default=None) is None
        True

        """
        try:
            return next(self.find_all(*tags, recursive=recursive))
        except StopIteration as e:
            if default is not _EMPTY_PARAM:
                return default

            msg = f"Unable to find tag by search criteria: {tags}"
            raise errors.SvgElementNotFoundError(msg) from e

    @override
    def _eq(self, other: Self) -> bool:
        return super()._eq(other) and all(
            c1 == c2
            for c1, c2 in zip(self.children, other.children, strict=True)
        )



from typing import Annotated, Literal, TypeAlias

import pydantic

from svglab import attrparse, models, utils


Unparsed: TypeAlias = str
"""Represents a type that is currently not parsed."""


# common string literals
All: TypeAlias = Literal["all"]
Auto: TypeAlias = Literal["auto"]
Inherit: TypeAlias = Literal["inherit"]
None_: TypeAlias = Literal["none"]

# basic types that are not parsed
AdvisoryTitle: TypeAlias = Unparsed
BeginValueList: TypeAlias = Unparsed
ClockValue: TypeAlias = Unparsed
ContentType: TypeAlias = Unparsed  # MIME type
Dasharray: TypeAlias = Unparsed  # list of percentages and lengths
EndValueList: TypeAlias = Unparsed
FamilyName: TypeAlias = Unparsed
FilterPrimitiveReference: TypeAlias = Unparsed
FuncIri: TypeAlias = Unparsed
GenericFamily: TypeAlias = Unparsed
Iri: TypeAlias = Unparsed
LanguageCodes: TypeAlias = Unparsed
LanguageId: TypeAlias = Unparsed
ListOfCoordinates: TypeAlias = Unparsed
ListOfLengths: TypeAlias = Unparsed
MediaDescriptors: TypeAlias = Unparsed
Name: TypeAlias = Unparsed
Percentage: TypeAlias = Unparsed
ProfileName: TypeAlias = Unparsed
Shape: TypeAlias = Unparsed
Urange: TypeAlias = Unparsed
XmlName: TypeAlias = Unparsed

# basic types that are parsed
AbsoluteSize: TypeAlias = Literal[
    "xx-small",
    "x-small",
    "small",
    "medium",
    "large",
    "x-large",
    "xx-large",
]
Angle: TypeAlias = attrparse.AngleType
Anything: TypeAlias = str
Boolean: TypeAlias = bool
Character: TypeAlias = Annotated[
    str, pydantic.Field(min_length=1, max_length=1)
]
Color: TypeAlias = attrparse.ColorType
Integer: TypeAlias = int
Length: TypeAlias = attrparse.LengthType
ListOfPoints: TypeAlias = attrparse.PointsType
Number: TypeAlias = float
PathData: TypeAlias = attrparse.DType
RelativeSize: TypeAlias = Literal["smaller", "larger"]
TransformList: TypeAlias = attrparse.TransformType

# composite types
Coordinate: TypeAlias = Length
CursorValue: TypeAlias = (
    FuncIri
    | Auto
    | Literal[
        "crosshair",
        "default",
        "pointer",
        "move",
        "e-resize",
        "ne-resize",
        "nw-resize",
        "n-resize",
        "se-resize",
        "sw-resize",
        "s-resize",
        "w-resize",
        "text",
        "wait",
        "help",
    ]
)
ListOfExtensions: TypeAlias = models.List[Iri]
ListOfFeatures: TypeAlias = models.List[Unparsed]
ListOfNames: TypeAlias = models.List[Name]
ListOfNumbers: TypeAlias = models.List[Number]
ListOfStrings: TypeAlias = models.List[Anything]
Miterlimit: TypeAlias = Annotated[Number, pydantic.Field(ge=1)]
NumericValue: TypeAlias = Number
NumberOptionalNumber: TypeAlias = Number | models.Tuple2[Number, Number]
OpacityValue: TypeAlias = Annotated[
    Number,
    pydantic.AfterValidator(
        lambda x: utils.clamp(x, min_value=0, max_value=1)
    ),
]
Paint: TypeAlias = (
    None_
    | Literal["currentColor"]
    | Inherit
    | Color
    # <funciri> [ none | currentColor | <color> [<icccolor>] ]
    | Unparsed
)



# ruff: noqa: N815

from typing import Literal

import pydantic

from svglab import constants, models
from svglab.attrs import common, types


class AccentHeight(common.Attr):
    accent_height: models.Attr[types.Number] = None


class Accumulate(common.Attr):
    accumulate: models.Attr[types.None_ | Literal["sum"]] = None


class Additive(common.Attr):
    additive: models.Attr[Literal["replace", "sum"]] = None


class Alphabetic(common.Attr):
    alphabetic: models.Attr[types.Number] = None


class Amplitude(common.Attr):
    amplitude: models.Attr[types.Number] = None


class ArabicForm(common.Attr):
    arabic_form: models.Attr[
        Literal["initial", "medial", "terminal", "isolated"]
    ] = None


class Ascent(common.Attr):
    ascent: models.Attr[types.Number] = None


class AttributeName(common.Attr):
    attributeName: models.Attr[types.Unparsed] = None


class AttributeType(common.Attr):
    attributeType: models.Attr[Literal["CSS", "XML"] | types.Auto] = None


class Azimuth(common.Attr):
    azimuth: models.Attr[types.Number] = None


class BaseFrequency(common.Attr):
    baseFrequency: models.Attr[types.NumberOptionalNumber] = None


class BaseProfile(common.Attr):
    baseProfile: models.Attr[types.ProfileName] = None


class Bbox(common.Attr):
    bbox: models.Attr[
        models.Tuple4[
            types.Number, types.Number, types.Number, types.Number
        ]
    ] = None


class Begin(common.Attr):
    begin: models.Attr[types.BeginValueList] = None


class Bias(common.Attr):
    bias: models.Attr[types.Number] = None


class By(common.Attr):
    by: models.Attr[types.Unparsed] = None


class CalcMode(common.Attr):
    calcMode: models.Attr[
        Literal["discrete", "linear", "paced", "spline"]
    ] = None


class CapHeight(common.Attr):
    cap_height: models.Attr[types.Number] = None


class Class(common.Attr):
    class_: models.Attr[types.ListOfStrings] = None


class ClipPathUnits(common.Attr):
    clipPathUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class ContentScriptType(common.Attr):
    contentScriptType: models.Attr[types.ContentType] = None


class ContentStyleType(common.Attr):
    contentStyleType: models.Attr[types.ContentType] = None


class Cx(common.Attr):
    cx: models.Attr[types.Coordinate] = None


class Cy(common.Attr):
    cy: models.Attr[types.Coordinate] = None


class D(common.Attr):
    d: models.Attr[types.PathData] = None


class Descent(common.Attr):
    descent: models.Attr[types.Number] = None


class DiffuseConstant(common.Attr):
    diffuseConstant: models.Attr[types.Number] = None


class Divisor(common.Attr):
    divisor: models.Attr[types.Number] = None


class Dur(common.Attr):
    dur: models.Attr[types.ClockValue | Literal["media", "indefinite"]] = (
        None
    )


class DxListOfLengths(common.Attr):
    dx: models.Attr[types.ListOfLengths] = None


class DxNumber(common.Attr):
    dx: models.Attr[types.Number] = None


class DyListOfLengths(common.Attr):
    dy: models.Attr[types.ListOfLengths] = None


class DyNumber(common.Attr):
    dy: models.Attr[types.Number] = None


class EdgeMode(common.Attr):
    edgeMode: models.Attr[Literal["duplicate", "wrap", "none"]] = None


class Elevation(common.Attr):
    elevation: models.Attr[types.Number] = None


class End(common.Attr):
    end: models.Attr[types.EndValueList] = None


class Exponent(common.Attr):
    exponent: models.Attr[types.Number] = None


class ExternalResourcesRequired(common.Attr):
    externalResourcesRequired: models.Attr[types.Boolean] = None


class FilterRes(common.Attr):
    filterRes: models.Attr[types.NumberOptionalNumber] = None


class FilterUnits(common.Attr):
    filterUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class FontFamily(common.Attr):
    font_family: models.Attr[
        models.List[types.FamilyName | types.GenericFamily]
    ] = None


class FontSize(common.Attr):
    font_size: models.Attr[types.All | types.ListOfLengths] = None


class Format(common.Attr):
    format: models.Attr[types.Unparsed] = None


class From(common.Attr):
    from_: models.Attr[types.Unparsed] = None


class Fx(common.Attr):
    fx: models.Attr[types.Coordinate] = None


class Fy(common.Attr):
    fy: models.Attr[types.Coordinate] = None


class G1(common.Attr):
    g1: models.Attr[types.ListOfNames] = None


class G2(common.Attr):
    g2: models.Attr[types.ListOfNames] = None


class GlyphName(common.Attr):
    glyph_name: models.Attr[types.ListOfNames] = None


class GlyphRef(common.Attr):
    glyphRef: models.Attr[types.Unparsed] = None


class GradientTransform(common.Attr):
    gradientTransform: models.Attr[types.TransformList] = None


class GradientUnits(common.Attr):
    gradientUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class Hanging(common.Attr):
    hanging: models.Attr[types.Number] = None


class Height(common.Attr):
    height: models.Attr[types.Length] = None


class HorizAdvX(common.Attr):
    horiz_adv_x: models.Attr[types.Number] = None


class HorizOriginX(common.Attr):
    horiz_origin_x: models.Attr[types.Number] = None


class HorizOriginY(common.Attr):
    horiz_origin_y: models.Attr[types.Number] = None


class Id(common.Attr):
    id: models.Attr[types.Name] = None


class Ideographic(common.Attr):
    ideographic: models.Attr[types.Number] = None


class In(common.Attr):
    in_: models.Attr[
        types.Paint | types.FilterPrimitiveReference | types.Unparsed
    ] = None


class In2(common.Attr):
    in2: models.Attr[
        types.Paint | types.FilterPrimitiveReference | types.Unparsed
    ] = None


class Intercept(common.Attr):
    intercept: models.Attr[types.Number] = None


class K(common.Attr):
    k: models.Attr[types.Number] = None


class K1(common.Attr):
    k1: models.Attr[types.Number] = None


class K2(common.Attr):
    k2: models.Attr[types.Number] = None


class K3(common.Attr):
    k3: models.Attr[types.Number] = None


class K4(common.Attr):
    k4: models.Attr[types.Number] = None


class KernelMatrix(common.Attr):
    kernelMatrix: models.Attr[types.ListOfNumbers] = None


class KernelUnitLength(common.Attr):
    kernelUnitLength: models.Attr[types.NumberOptionalNumber] = None


class KeyPoints(common.Attr):
    keyPoints: models.Attr[types.ListOfNumbers] = None


class KeySplines(common.Attr):
    keySplines: models.Attr[types.Unparsed] = None


class KeyTimes(common.Attr):
    keyTimes: models.Attr[types.Unparsed] = None


class Lang(common.Attr):
    lang: models.Attr[types.LanguageCodes] = None


class LengthAdjust(common.Attr):
    lengthAdjust: models.Attr[Literal["spacing", "spacingAndGlyphs"]] = (
        None
    )


class LimitingConeAngle(common.Attr):
    limitingConeAngle: models.Attr[types.Number] = None


class Local(common.Attr):
    local: models.Attr[types.Unparsed] = None


class MarkerHeight(common.Attr):
    markerHeight: models.Attr[types.Length] = None


class MarkerUnits(common.Attr):
    markerUnits: models.Attr[Literal["strokeWidth", "userSpaceOnUse"]] = (
        None
    )


class MarkerWidth(common.Attr):
    markerWidth: models.Attr[types.Length] = None


class MaskContentUnits(common.Attr):
    maskContentUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class MaskUnits(common.Attr):
    maskUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class Mathematical(common.Attr):
    mathematical: models.Attr[types.Number] = None


class Max(common.Attr):
    max: models.Attr[Literal["media"] | types.ClockValue] = None


class Media(common.Attr):
    media: models.Attr[types.MediaDescriptors] = None


class Method(common.Attr):
    method: models.Attr[Literal["align", "stretch"]] = None


class Min(common.Attr):
    min: models.Attr[Literal["media"] | types.ClockValue] = None


class Mode(common.Attr):
    mode: models.Attr[
        Literal["normal", "multiply", "screen", "darken", "lighten"]
    ] = None


class NameName(common.Attr):
    name: models.Attr[types.Name] = None


class NameAnything(common.Attr):
    name: models.Attr[types.Anything] = None


class NumOctaves(common.Attr):
    numOctaves: models.Attr[types.Integer] = None


class OffsetNumberPercentage(common.Attr):
    offset: models.Attr[types.Number | types.Percentage] = None


class OffsetNumber(common.Attr):
    offset: models.Attr[types.Number] = None


class OnAbort(common.Attr):
    onabort: models.Attr[types.Anything] = None


class OnActivate(common.Attr):
    onactivate: models.Attr[types.Anything] = None


class OnBegin(common.Attr):
    onbegin: models.Attr[types.Anything] = None


class OnClick(common.Attr):
    onclick: models.Attr[types.Anything] = None


class OnEnd(common.Attr):
    onend: models.Attr[types.Anything] = None


class OnError(common.Attr):
    onerror: models.Attr[types.Anything] = None


class OnFocusIn(common.Attr):
    onfocusin: models.Attr[types.Anything] = None


class OnFocusOut(common.Attr):
    onfocusout: models.Attr[types.Anything] = None


class OnLoad(common.Attr):
    onload: models.Attr[types.Anything] = None


class OnMouseDown(common.Attr):
    onmousedown: models.Attr[types.Anything] = None


class OnMouseMove(common.Attr):
    onmousemove: models.Attr[types.Anything] = None


class OnMouseOut(common.Attr):
    onmouseout: models.Attr[types.Anything] = None


class OnMouseOver(common.Attr):
    onmouseover: models.Attr[types.Anything] = None


class OnMouseUp(common.Attr):
    onmouseup: models.Attr[types.Anything] = None


class OnRepeat(common.Attr):
    onrepeat: models.Attr[types.Anything] = None


class OnResize(common.Attr):
    onresize: models.Attr[types.Anything] = None


class OnScroll(common.Attr):
    onscroll: models.Attr[types.Anything] = None


class OnUnload(common.Attr):
    onunload: models.Attr[types.Anything] = None


class OnZoom(common.Attr):
    onzoom: models.Attr[types.Anything] = None


class OperatorFeComposite(common.Attr):
    operator: models.Attr[
        Literal["over", "in", "out", "atop", "xor", "arithmetic"]
    ] = None


class OperatorFeMorphology(common.Attr):
    operator: models.Attr[Literal["erode", "dilate"]] = None


class Order(common.Attr):
    order: models.Attr[types.NumberOptionalNumber] = None


class Orient(common.Attr):
    orient: models.Attr[types.Auto | types.Angle] = None


class Orientation(common.Attr):
    orientation: models.Attr[Literal["h", "v"]] = None


class Origin(common.Attr):
    origin: models.Attr[Literal["default"]] = None


class OverlinePosition(common.Attr):
    overline_position: models.Attr[types.Number] = None


class OverlineThickness(common.Attr):
    overline_thickness: models.Attr[types.Number] = None


class Panose1(common.Attr):
    panose1: models.Attr[
        models.Tuple[
            tuple[
                types.Integer,
                types.Integer,
                types.Integer,
                types.Integer,
                types.Integer,
                types.Integer,
                types.Integer,
                types.Integer,
                types.Integer,
                types.Integer,
            ]
        ]
    ] = None


class Path(common.Attr):
    path: models.Attr[types.PathData] = None


class PathLength(common.Attr):
    pathLength: models.Attr[types.Number] = None


class PatternContentUnits(common.Attr):
    patternContentUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class PatternTransform(common.Attr):
    patternTransform: models.Attr[types.TransformList] = None


class PatternUnits(common.Attr):
    patternUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class Points(common.Attr):
    points: models.Attr[types.ListOfPoints] = None


class PointsAtX(common.Attr):
    pointsAtX: models.Attr[types.Number] = None


class PointsAtY(common.Attr):
    pointsAtY: models.Attr[types.Number] = None


class PointsAtZ(common.Attr):
    pointsAtZ: models.Attr[types.Number] = None


class PreserveAlpha(common.Attr):
    preserveAlpha: models.Attr[types.Boolean] = None


class PreserveAspectRatio(common.Attr):
    preserveAspectRatio: models.Attr[
        None
        | Literal[
            "xMinYMin",
            "xMidYMin",
            "xMaxYMin",
            "xMinYMid",
            "xMidYMid",
            "xMaxYMid",
            "xMinYMax",
            "xMidYMax",
            "xMaxYMax",
        ]
        | types.Unparsed
    ] = None


class PrimitiveUnits(common.Attr):
    primitiveUnits: models.Attr[
        Literal["userSpaceOnUse", "objectBoundingBox"]
    ] = None


class R(common.Attr):
    r: models.Attr[types.Length] = None


class Radius(common.Attr):
    radius: models.Attr[types.NumberOptionalNumber] = None


class RefX(common.Attr):
    ref_x: models.Attr[types.Coordinate] = None


class RefY(common.Attr):
    ref_y: models.Attr[types.Coordinate] = None


class RenderingIntent(common.Attr):
    rendering_intent: models.Attr[
        Literal[
            "auto",
            "perceptual",
            "relative-colorimetric",
            "saturation",
            "absolute-colorimetric",
        ]
    ] = None


class RepeatCount(common.Attr):
    repeatCount: models.Attr[
        Literal["indefinite"] | types.NumericValue
    ] = None


class RepeatDur(common.Attr):
    repeatDur: models.Attr[Literal["indefinite"] | types.ClockValue] = None


class RequiredExtensions(common.Attr):
    requiredExtensions: models.Attr[types.ListOfExtensions] = None


class RequiredFeatures(common.Attr):
    requiredFeatures: models.Attr[types.ListOfFeatures] = None


class Restart(common.Attr):
    restart: models.Attr[Literal["always", "whenNotActive", "never"]] = (
        None
    )


class Result(common.Attr):
    result: models.Attr[types.FilterPrimitiveReference] = None


class RotateListOfNumbers(common.Attr):
    rotate: models.Attr[types.ListOfNumbers] = None


class RotateNumberAutoAutoReverse(common.Attr):
    rotate: (
        models.Attr[types.Number] | types.Auto | Literal["auto-reverse"]
    ) = None


class Rx(common.Attr):
    rx: models.Attr[types.Length] = None


class Ry(common.Attr):
    ry: models.Attr[types.Length] = None


class Scale(common.Attr):
    scale: models.Attr[types.Number] = None


class Seed(common.Attr):
    seed: models.Attr[types.Number] = None


class Slope(common.Attr):
    slope: models.Attr[types.Number] = None


class Spacing(common.Attr):
    spacing: models.Attr[Literal["auto", "exact"]] = None


class SpecularConstant(common.Attr):
    specularConstant: models.Attr[types.Number] = None


class SpecularExponent(common.Attr):
    specularExponent: models.Attr[types.Number] = None


class SpreadMethod(common.Attr):
    spreadMethod: models.Attr[Literal["pad", "reflect", "repeat"]] = None


class StartOffset(common.Attr):
    startOffset: models.Attr[types.Length] = None


class StdDeviation(common.Attr):
    stdDeviation: models.Attr[types.NumberOptionalNumber] = None


class Stemh(common.Attr):
    stemh: models.Attr[types.Number] = None


class Stemv(common.Attr):
    stemv: models.Attr[types.Number] = None


class StitchTiles(common.Attr):
    stitchTiles: models.Attr[Literal["stitch", "noStitch"]] = None


class StrikethroughPosition(common.Attr):
    strikethrough_position: models.Attr[types.Number] = None


class StrikethroughThickness(common.Attr):
    strikethrough_thickness: models.Attr[types.Number] = None


class String(common.Attr):
    string: models.Attr[types.Anything] = None


class Style(common.Attr):
    style: models.Attr[types.Unparsed] = None


class SurfaceScale(common.Attr):
    surfaceScale: models.Attr[types.Number] = None


class SystemLanguage(common.Attr):
    systemLanguage: models.Attr[types.LanguageCodes] = None


class TableValues(common.Attr):
    tableValues: models.Attr[types.ListOfNumbers] = None


class Target(common.Attr):
    target: models.Attr[
        Literal["_replace", "_self", "_parent", "_top", "_blank"]
        | types.XmlName
    ] = None


class TargetX(common.Attr):
    targetX: models.Attr[types.Integer] = None


class TargetY(common.Attr):
    targetY: models.Attr[types.Integer] = None


class TextLength(common.Attr):
    textLength: models.Attr[types.Length] = None


class Title(common.Attr):
    title: models.Attr[types.AdvisoryTitle] = None


class To(common.Attr):
    to: models.Attr[types.Unparsed] = None


class Transform(common.Attr):
    transform: models.Attr[types.TransformList] = None


class TypeAnimateTransform(common.Attr):
    type: models.Attr[
        Literal["translate", "scale", "rotate", "skewX", "skewY"]
    ] = None


class TypeFeColorMatrix(common.Attr):
    type: models.Attr[
        Literal["matrix", "saturate", "hueRotate", "luminanceToAlpha"]
    ] = None


class TypeFeTurbluence(common.Attr):
    type: models.Attr[Literal["fractalNoise", "turbulence"]] = None


class TypeContentType(common.Attr):
    type: models.Attr[types.ContentType] = None


class TypeFeFunc(common.Attr):
    type: models.Attr[
        Literal["identity", "table", "discrete", "linear", "gamma"]
    ] = None


class U1(common.Attr):
    u1: models.Attr[models.List[types.Character | types.Urange]] = None


class U2(common.Attr):
    u2: models.Attr[models.List[types.Character | types.Urange]] = None


class UnderlinePosition(common.Attr):
    underline_position: models.Attr[types.Number] = None


class UnderlineThickness(common.Attr):
    underline_thickness: models.Attr[types.Number] = None


class Unicode(common.Attr):
    unicode: models.Attr[types.Anything] = None


class UnicodeRange(common.Attr):
    unicode_range: models.Attr[models.List[types.Urange]] = None


class UnitsPerEm(common.Attr):
    units_per_em: models.Attr[types.Number] = None


class VAlphabetic(common.Attr):
    v_alphabetic: models.Attr[types.Number] = None


class VHanging(common.Attr):
    v_hanging: models.Attr[types.Number] = None


class VIdeographic(common.Attr):
    v_ideographic: models.Attr[types.Number] = None


class VMathematical(common.Attr):
    v_mathematical: models.Attr[types.Number] = None


class ValuesListOfNumbers(common.Attr):
    values: models.Attr[types.ListOfNumbers] = None


class ValuesList(common.Attr):
    values: models.Attr[types.Unparsed] = None


class Version(common.Attr):
    version: models.Attr[Literal["1.1"]] = None


class VertAdvY(common.Attr):
    vert_adv_y: models.Attr[types.Number] = None


class VertOriginX(common.Attr):
    vert_origin_x: models.Attr[types.Number] = None


class VertOriginY(common.Attr):
    vert_origin_y: models.Attr[types.Number] = None


class ViewBox(common.Attr):
    viewBox: models.Attr[
        models.Tuple4[
            types.Number, types.Number, types.Number, types.Number
        ]
    ] = None


class ViewTarget(common.Attr):
    viewTarget: models.Attr[models.List[types.XmlName]] = None


class Width(common.Attr):
    width: models.Attr[types.Length] = None


class Widths(common.Attr):
    widths: models.Attr[
        models.List[
            types.Urange | models.Tuple2[types.Urange, types.Number]
        ]
    ] = None


class XListOfCoordinates(common.Attr):
    x: models.Attr[types.ListOfCoordinates] = None


class XCoordinate(common.Attr):
    x: models.Attr[types.Coordinate] = None


class XNumber(common.Attr):
    x: models.Attr[types.Number] = None


class XHeight(common.Attr):
    x_height: models.Attr[types.Number] = None


class X1(common.Attr):
    x1: models.Attr[types.Coordinate] = None


class X2(common.Attr):
    x2: models.Attr[types.Coordinate] = None


class XChannelSelector(common.Attr):
    xChannelSelector: models.Attr[Literal["R", "G", "B", "A"]] = None


class XlinkActuateOnRequest(common.Attr):
    xlink_actuate: models.Attr[Literal["onRequest"]] = None


class XlinkActuateOnLoad(common.Attr):
    xlink_actuate: models.Attr[Literal["onLoad"]] = None


class XlinkArcrole(common.Attr):
    xlink_arcrole: models.Attr[types.Iri] = None


class XlinkHref(common.Attr):
    xlink_href: models.Attr[types.Iri] = None


class XlinkRole(common.Attr):
    xlink_role: models.Attr[types.Iri] = None


class XlinkShow(common.Attr):
    xlink_show: models.Attr[
        Literal["new", "replace", "embed", "other", "none"]
    ] = None


class XlinkTitle(common.Attr):
    xlink_title: models.Attr[types.Anything] = None


class XlinkType(common.Attr):
    xlink_type: models.Attr[Literal["simple"]] = None


class XmlBase(common.Attr):
    xml_base: models.Attr[types.Iri] = None


class XmlLang(common.Attr):
    xml_lang: models.Attr[types.LanguageId] = None


class Xmlns(common.Attr):
    # cannot use Final here because pydantic will consider the field a ClassVar
    # see https://github.com/pydantic/pydantic/issues/6253
    xmlns: constants.Xmlns = pydantic.Field(
        default=constants.DEFAULT_XMLNS, frozen=True
    )


class XmlSpace(common.Attr):
    xml_space: models.Attr[Literal["default", "preserve"]] = None


class YListOfCoordinates(common.Attr):
    y: models.Attr[types.ListOfCoordinates] = None


class YCoordinate(common.Attr):
    y: models.Attr[types.Coordinate] = None


class YNumber(common.Attr):
    y: models.Attr[types.Number] = None


class Y1(common.Attr):
    y1: models.Attr[types.Coordinate] = None


class Y2(common.Attr):
    y2: models.Attr[types.Coordinate] = None


class YChannelSelector(common.Attr):
    yChannelSelector: models.Attr[Literal["R", "G", "B", "A"]] = None


class Z(common.Attr):
    z: models.Attr[types.Number] = None


class ZoomAndPan(common.Attr):
    zoomAndPan: models.Attr[Literal["disable", "magnify"]] = None



from __future__ import annotations

from typing import Annotated, SupportsComplex, TypeAlias, final

import lark
import pydantic
from typing_extensions import override

from svglab import serialize
from svglab.attrparse import utils


@final
@pydantic.dataclasses.dataclass(frozen=True)
class Point(SupportsComplex, serialize.CustomSerializable):
    """A point in a 2D plane.

    Attributes:
        x: The x-coordinate of the point.
        y: The y-coordinate of the point.

    Examples:
        >>> point = Point(1, 2)
        >>> point.x
        1.0
        >>> point.y
        2.0
        >>> point == Point(1, 2)
        True
        >>> point == Point(1, 3)
        False
        >>> point += Point(2, 3)
        >>> point
        Point(x=3.0, y=5.0)
        >>> point -= Point(1, 2)
        >>> point
        Point(x=2.0, y=3.0)
        >>> point *= 2
        >>> point
        Point(x=4.0, y=6.0)
        >>> point /= 2
        >>> point
        Point(x=2.0, y=3.0)
        >>> point + point
        Point(x=4.0, y=6.0)

    """

    x: float
    y: float

    @override
    def serialize(self) -> str:
        formatter = serialize.get_current_formatter()
        x, y = serialize.serialize(self.x, self.y)

        return f"{x}{formatter.point_separator}{y}"

    def __add__(self, other: Point) -> Point:
        return Point(self.x + other.x, self.y + other.y)

    def __neg__(self) -> Point:
        return self * -1

    def __sub__(self, other: Point) -> Point:
        return self + -other

    def __mul__(self, scalar: float) -> Point:
        return Point(self.x * scalar, self.y * scalar)

    def __truediv__(self, scalar: float) -> Point:
        return Point(self.x / scalar, self.y / scalar)

    def __complex__(self) -> complex:
        return complex(self.x, self.y)

    @classmethod
    def from_complex(cls, value: complex, /) -> Point:
        return cls(value.real, value.imag)

    @classmethod
    def zero(cls) -> Point:
        return cls(0, 0)


@lark.v_args(inline=True)
class _Transformer(lark.Transformer[object, Point]):
    number = float
    point = Point


PointType: TypeAlias = Annotated[
    Point,
    utils.get_validator(grammar="point.lark", transformer=_Transformer()),
]



from svglab.attrs import groups
from svglab.elements import common


class Element(groups.Core, groups.Presentation, common.Tag):
    pass


class GraphicsElement(
    groups.GraphicalEvents,  # TODO: check if this is correct
    Element,
):
    pass


class Shape(GraphicsElement):
    pass


class BasicShape(Shape):
    pass


class AnimationElement(
    groups.AnimationEvents, groups.AnimationTiming, Element
):
    pass


class ContainerElement(
    groups.GraphicalEvents,  # TODO: check if this is correct
    common.PairedTag,
):
    pass


class DescriptiveElement(Element):
    pass


class FilterPrimitiveElement(groups.FilterPrimitives, Element):
    pass


class GradientElement(Element):
    pass


class GraphicsReferencingElement(Element):
    pass


class LightSourceElement(Element):
    pass


class StructuralElement(Element):
    pass


class TextContentElement(GraphicsElement):
    pass


class TextContentChildElement(Element):
    pass


class TextContentBlockElement(Element):
    pass



from typing import final

import bs4
from typing_extensions import override

from svglab.elements import common


@final
class CData(common.TextElement):
    """A `CDATA` section.

    A `CDATA` section is a block of text that is not parsed by the XML parser,
    but is interpreted verbatim.

    `CDATA` sections are used to include text that contains characters
    that would otherwise be interpreted as XML markup.

    Example: `<![CDATA[<g id="foo"></g>]]>`
    """

    def __init__(self, content: str, /) -> None:
        super().__init__(content=content)

    @override
    def to_beautifulsoup_object(self) -> bs4.CData:
        return bs4.CData(self.content)


@final
class Comment(common.TextElement):
    """A comment.

    A comment is a block of text that is not parsed by the XML parser,
    but is ignored.

    Comments are used to include notes and other information that is not
    intended to be displayed to the user.

    Example: `<!-- This is a comment -->`
    """

    def __init__(self, content: str, /) -> None:
        super().__init__(content=content)

    @override
    def to_beautifulsoup_object(self) -> bs4.Comment:
        return bs4.Comment(self.content)


@final
class RawText(common.TextElement):
    """A text node.

    A text node is a block of text that is parsed by the XML parser.

    Text nodes are used to include text that is intended to be displayed
    to the user.

    Example: `Hello, world!`
    """

    def __init__(self, content: str, /) -> None:
        super().__init__(content=content)

    @override
    def to_beautifulsoup_object(self) -> bs4.NavigableString:
        return bs4.NavigableString(self.content)



from __future__ import annotations

from typing import Annotated, Literal, TypeAlias

import lark
import pydantic
from typing_extensions import override

from svglab import serialize
from svglab.attrparse import utils


AngleUnit: TypeAlias = Literal["deg", "grad", "rad"]


@pydantic.dataclasses.dataclass
class Angle(serialize.CustomSerializable):
    """Represents the SVG `<angle>` type.

    An angle is a number optionally followed by a unit. Available units are:

    - `deg`: degrees
    - `grad`: gradians
    - `rad`: radians
    """

    value: float
    unit: AngleUnit | None = "deg"

    @override
    def serialize(self) -> str:
        value = serialize.serialize(self.value)
        return f"{value}{self.unit or ''}"


@lark.v_args(inline=True)
class _Transformer(lark.Transformer[object, Angle]):
    number = float
    angle = Angle


AngleType: TypeAlias = Annotated[
    Angle,
    utils.get_validator(grammar="angle.lark", transformer=_Transformer()),
]



import keyword
import typing
from typing import Final, Literal, TypeAlias

import bidict


AttributeName: TypeAlias = Literal[
    "accent-height",
    "accumulate",
    "additive",
    "alignment-baseline",
    "alphabetic",
    "amplitude",
    "arabic-form",
    "ascent",
    "attributeName",
    "attributeType",
    "azimuth",
    "baseFrequency",
    "baseline-shift",
    "baseProfile",
    "bbox",
    "begin",
    "bias",
    "by",
    "calcMode",
    "cap-height",
    "class",
    "clip-path",
    "clip-rule",
    "clip",
    "clipPathUnits",
    "color-interpolation-filters",
    "color-interpolation",
    "color-profile",
    "color-rendering",
    "color",
    "contentScriptType",
    "contentStyleType",
    "cursor",
    "cx",
    "cy",
    "d",
    "descent",
    "diffuseConstant",
    "direction",
    "display",
    "divisor",
    "dominant-baseline",
    "dur",
    "dx",
    "dy",
    "edgeMode",
    "elevation",
    "enable-background",
    "end",
    "exponent",
    "externalResourcesRequired",
    "fill-opacity",
    "fill-rule",
    "fill",
    "filter",
    "filterRes",
    "filterUnits",
    "flood-color",
    "flood-opacity",
    "font-family",
    "font-size-adjust",
    "font-size",
    "font-stretch",
    "font-style",
    "font-variant",
    "font-weight",
    "format",
    "from",
    "fx",
    "fy",
    "g1",
    "g2",
    "glyph-name",
    "glyph-orientation-horizontal",
    "glyph-orientation-vertical",
    "glyphRef",
    "gradientTransform",
    "gradientUnits",
    "hanging",
    "height",
    "horiz-adv-x",
    "horiz-origin-x",
    "horiz-origin-y",
    "id",
    "ideographic",
    "image-rendering",
    "in",
    "in2",
    "intercept",
    "k",
    "k1",
    "k2",
    "k3",
    "k4",
    "kernelMatrix",
    "kernelUnitLength",
    "kerning",
    "keyPoints",
    "keySplines",
    "keyTimes",
    "lang",
    "lengthAdjust",
    "letter-spacing",
    "lighting-color",
    "limitingConeAngle",
    "local",
    "marker-end",
    "marker-mid",
    "marker-start",
    "markerHeight",
    "markerUnits",
    "markerWidth",
    "mask",
    "maskContentUnits",
    "maskUnits",
    "mathematical",
    "max",
    "media",
    "method",
    "min",
    "mode",
    "name",
    "numOctaves",
    "offset",
    "opacity",
    "operator",
    "order",
    "orient",
    "orientation",
    "origin",
    "overflow",
    "overline-position",
    "overline-thickness",
    "panose-1",
    "path",
    "pathLength",
    "patternContentUnits",
    "patternTransform",
    "patternUnits",
    "pointer-events",
    "points",
    "pointsAtX",
    "pointsAtY",
    "pointsAtZ",
    "preserveAlpha",
    "preserveAspectRatio",
    "primitiveUnits",
    "r",
    "radius",
    "refX",
    "refY",
    "rendering-intent",
    "repeatCount",
    "repeatDur",
    "requiredExtensions",
    "requiredFeatures",
    "restart",
    "result",
    "rotate",
    "rx",
    "ry",
    "scale",
    "seed",
    "shape-rendering",
    "slope",
    "spacing",
    "specularConstant",
    "specularExponent",
    "spreadMethod",
    "startOffset",
    "stdDeviation",
    "stemh",
    "stemv",
    "stitchTiles",
    "stop-color",
    "stop-opacity",
    "strikethrough-position",
    "strikethrough-thickness",
    "string",
    "stroke-dasharray",
    "stroke-dashoffset",
    "stroke-linecap",
    "stroke-linejoin",
    "stroke-miterlimit",
    "stroke-opacity",
    "stroke-width",
    "stroke",
    "style",
    "surfaceScale",
    "systemLanguage",
    "tableValues",
    "target",
    "targetX",
    "targetY",
    "text-anchor",
    "text-decoration",
    "text-rendering",
    "textLength",
    "title",
    "to",
    "transform",
    "type",
    "u1",
    "u2",
    "underline-position",
    "underline-thickness",
    "unicode-bidi",
    "unicode-range",
    "unicode",
    "units-per-em",
    "v-alphabetic",
    "v-hanging",
    "v-ideographic",
    "v-mathematical",
    "values",
    "version",
    "vert-adv-y",
    "vert-origin-x",
    "vert-origin-y",
    "viewBox",
    "viewTarget",
    "visibility",
    "width",
    "widths",
    "word-spacing",
    "writing-mode",
    "x-height",
    "x",
    "x1",
    "x2",
    "xChannelSelector",
    "xlink:actuate",
    "xlink:arcrole",
    "xlink:href",
    "xlink:role",
    "xlink:show",
    "xlink:title",
    "xlink:type",
    "xml:base",
    "xml:lang",
    "xml:space",
    "xmlns",
    "y",
    "y1",
    "y2",
    "yChannelSelector",
    "z",
    "zoomAndPan",
]
""" Type for all SVG attribute names. """


def _is_valid_identifier(name: str) -> bool:
    """Check if a string is a valid identifier and not a reserved keyword.

    Args:
        name: The string to check.

    Returns:
        `True` if the string is a valid Python identifier, `False` otherwise.

    Examples:
        >>> _is_valid_identifier("foo")
        True
        >>> _is_valid_identifier("class")
        False
        >>> _is_valid_identifier("123")
        False
        >>> _is_valid_identifier("for")
        False

    """
    return name.isidentifier() and not keyword.iskeyword(name)


def _normalize_attr_name(name: AttributeName) -> str:
    """Convert an SVG attribute name to a valid Python identifier.

    Args:
        name: The attribute name to normalize.

    Returns:
        The normalized attribute name.

    Raises:
        ValueError: If the attribute name cannot be normalized.

    Examples:
        >>> _normalize_attr_name("width")
        'width'
        >>> _normalize_attr_name("stroke-width")
        'stroke_width'
        >>> _normalize_attr_name("xlink:href")
        'xlink_href'
        >>> _normalize_attr_name("class")
        'class_'

    """
    normalized: str = name

    substitutions = {"-": "_", ":": "_"}

    for old, new in substitutions.items():
        normalized = normalized.replace(old, new)

    if not _is_valid_identifier(normalized):
        normalized = f"{normalized}_"

    if not _is_valid_identifier(normalized):
        msg = f"Cannot normalize attribute name: {name!r}"
        raise ValueError(msg)

    return normalized


_ATTRIBUTE_NAMES: Final[frozenset[AttributeName]] = frozenset(
    typing.get_args(AttributeName)
)
"""A set of all valid SVG attribute names."""


ATTR_NAME_TO_NORMALIZED: Final = bidict.frozenbidict(
    {attr: _normalize_attr_name(attr) for attr in _ATTRIBUTE_NAMES}
)
"""
A bidirectional mapping of SVG attribute names to normalized
Python identifiers.
"""



import itertools
import pathlib
from typing import Final, TypeVar

import lark
import pydantic
from typing_extensions import LiteralString


_T = TypeVar("_T")
_Leaf_T = TypeVar("_Leaf_T")
_Return_T = TypeVar("_Return_T")


_CURRENT_DIR: Final = pathlib.Path(__file__).parent
_GRAMMARS_DIR: Final = _CURRENT_DIR / "grammars"


def _parse(
    text: str,
    /,
    *,
    grammar: LiteralString,
    transformer: lark.Transformer[_Leaf_T, _Return_T],
    **kwargs: object,
) -> lark.ParseTree:
    parser = lark.Lark.open(
        grammar_filename=str(_GRAMMARS_DIR / grammar),
        rel_to=None,
        cache=True,
        maybe_placeholders=False,
        ordered_sets=False,
        parser="lalr",
        propagate_positions=False,
        transformer=transformer,
        **kwargs,
    )

    try:
        return parser.parse(text)
    except lark.LarkError as e:
        msg = f"Failed to parse text with grammar {grammar!r}. Reason: \n\n{e}"
        raise ValueError(msg) from e


def get_validator(
    *,
    grammar: LiteralString,
    transformer: lark.Transformer[_Leaf_T, _Return_T],
    **kwargs: object,
) -> pydantic.BeforeValidator:
    def validator(value: object) -> object:
        if isinstance(value, str):
            return _parse(
                value, grammar=grammar, transformer=transformer, **kwargs
            )

        return value

    return pydantic.BeforeValidator(validator)


def v_args_to_list(*values: _T) -> list[_T]:
    # drop the first value, which is the transformer instance (self)
    return list(itertools.islice(values, 1, None))



from typing import Annotated, TypeAlias

import lark
import lark.visitors

from svglab import attrparse
from svglab.attrparse import utils


Points: TypeAlias = list[attrparse.Point]


@lark.v_args(inline=True)
class _Transformer(lark.Transformer[object, Points]):
    number = float
    point = attrparse.Point

    points = utils.v_args_to_list


PointsType: TypeAlias = Annotated[
    Points,
    utils.get_validator(grammar="points.lark", transformer=_Transformer()),
]



from __future__ import annotations

from collections.abc import Generator, Iterable, Iterator, MutableSequence
from typing import (
    Final,
    Literal,
    SupportsIndex,
    TypeAlias,
    final,
    overload,
)

import pydantic
import pydantic_core
import svgpathtools
from typing_extensions import Self, override

from svglab import models, serialize, utils
from svglab.attrparse import point


_AbsolutePathCommandChar: TypeAlias = Literal[
    "M", "L", "H", "V", "C", "S", "Q", "T", "A", "Z"
]


@final
@pydantic.dataclasses.dataclass
class MoveTo:
    end: point.Point

    def __add__(self, other: point.Point, /) -> Self:
        return type(self)(end=self.end + other)

    def __sub__(self, other: point.Point, /) -> Self:
        return type(self)(end=self.end - other)


@pydantic.dataclasses.dataclass
class LineTo:
    end: point.Point

    def __add__(self, other: point.Point, /) -> Self:
        return type(self)(end=self.end + other)

    def __sub__(self, other: point.Point, /) -> Self:
        return type(self)(end=self.end - other)


@final
@pydantic.dataclasses.dataclass
class QuadraticBezierTo:
    control: point.Point
    end: point.Point

    def __add__(self, other: point.Point, /) -> Self:
        return type(self)(
            control=self.control + other, end=self.end + other
        )

    def __sub__(self, other: point.Point, /) -> Self:
        return type(self)(
            control=self.control - other, end=self.end - other
        )


@final
@pydantic.dataclasses.dataclass
class CubicBezierTo:
    control1: point.Point
    control2: point.Point
    end: point.Point

    def __add__(self, other: point.Point, /) -> Self:
        return type(self)(
            control1=self.control1 + other,
            control2=self.control2 + other,
            end=self.end + other,
        )

    def __sub__(self, other: point.Point, /) -> Self:
        return type(self)(
            control1=self.control1 - other,
            control2=self.control2 - other,
            end=self.end - other,
        )


@final
@pydantic.dataclasses.dataclass
class ArcTo:
    radius: point.Point
    angle: float
    large: bool
    sweep: bool
    end: point.Point

    def __add__(self, other: point.Point, /) -> Self:
        return type(self)(
            radius=self.radius + other,
            angle=self.angle,
            large=self.large,
            sweep=self.sweep,
            end=self.end + other,
        )

    def __sub__(self, other: point.Point, /) -> Self:
        return type(self)(
            radius=self.radius - other,
            angle=self.angle,
            large=self.large,
            sweep=self.sweep,
            end=self.end - other,
        )


@final
@pydantic.dataclasses.dataclass
class ClosePath(LineTo):
    pass


PathCommand: TypeAlias = (
    MoveTo | LineTo | QuadraticBezierTo | CubicBezierTo | ArcTo | ClosePath
)


@final
class D(
    MutableSequence[PathCommand],
    models.CustomModel,
    serialize.CustomSerializable,
):
    def __init__(
        self, iterable: Iterable[PathCommand] | None = None, /
    ) -> None:
        self.__commands: Final[list[PathCommand]] = list(iterable or [])

    @override
    def __len__(self) -> int:
        return len(self.__commands)

    @overload
    def __getitem__(self, index: int) -> PathCommand: ...

    @overload
    def __getitem__(self, index: slice) -> Self: ...

    @override
    def __getitem__(self, index: int | slice) -> PathCommand | Self:
        if isinstance(index, SupportsIndex):
            return self.__commands[index]

        return type(self)(self.__commands[index])

    @overload
    def __delitem__(self, index: int) -> None: ...

    @overload
    def __delitem__(self, index: slice) -> None: ...

    @override
    def __delitem__(self, index: int | slice) -> None:
        if isinstance(index, SupportsIndex):
            del self.__commands[index]
        else:
            del self.__commands[index]

    @overload
    def __setitem__(self, index: int, values: PathCommand) -> None: ...

    @overload
    def __setitem__(
        self, index: slice, values: Iterable[PathCommand]
    ) -> None: ...

    @override
    def __setitem__(
        self,
        index: int | slice,
        values: PathCommand | Iterable[PathCommand],
    ) -> None:
        if isinstance(index, SupportsIndex):
            assert isinstance(values, PathCommand)
            self.__commands[index] = values
        else:
            assert isinstance(values, Iterable)
            self.__commands[index] = values

    @override
    def insert(self, index: SupportsIndex, value: PathCommand) -> None:
        self.__commands.insert(index, value)

    @property
    def start(self) -> point.Point:
        return next(
            (cmd.end for cmd in self if not isinstance(cmd, MoveTo)),
            point.Point.zero(),
        )

    @property
    def end(self) -> point.Point:
        if not self:
            return point.Point.zero()

        return self[-1].end

    def __add(
        self, command: PathCommand, *, relative: bool = False
    ) -> Self:
        if relative:
            command += self.end

        self.append(command)

        return self

    def move_to(
        self, end: point.Point, /, *, relative: bool = False
    ) -> Self:
        return self.__add(MoveTo(end=end), relative=relative)

    def line_to(
        self, end: point.Point, /, *, relative: bool = False
    ) -> Self:
        return self.__add(LineTo(end=end), relative=relative)

    def quadratic_bezier_to(
        self,
        control: point.Point,
        end: point.Point,
        *,
        relative: bool = False,
    ) -> Self:
        return self.__add(
            QuadraticBezierTo(control=control, end=end), relative=relative
        )

    def cubic_bezier_to(
        self,
        control1: point.Point,
        control2: point.Point,
        end: point.Point,
        *,
        relative: bool = False,
    ) -> Self:
        return self.__add(
            CubicBezierTo(control1=control1, control2=control2, end=end),
            relative=relative,
        )

    def arc_to(  # noqa: PLR0913
        self,
        radius: point.Point,
        angle: float,
        end: point.Point,
        *,
        large: bool,
        sweep: bool,
        relative: bool = False,
    ) -> Self:
        return self.__add(
            ArcTo(
                radius=radius,
                angle=angle,
                large=large,
                sweep=sweep,
                end=end,
            ),
            relative=relative,
        )

    def close(self) -> Self:
        last_continuous_subpath = utils.take_last(
            self.__continuous_subpaths()
        )

        end = (
            self.start
            if last_continuous_subpath is None
            else last_continuous_subpath.start
        )

        return self.__add(ClosePath(end=end))

    def is_continuous(self) -> bool:
        """Determine whether the path is continuous.

        A continuous path is a path that does not contain any `MoveTo`
        commands in between other commands. A continuous path may contain
        leading or trailing `MoveTo` commands.

        Returns:
            `True` if the path is continuous, `False` otherwise.

        Examples:
            >>> d = D()
            >>> d.is_continuous()
            True
            >>> _ = d.move_to(point.Point(0, 0)).line_to(point.Point(1, 1))
            >>> d.is_continuous()
            True
            >>> _ = d.move_to(point.Point(2, 2)).line_to(point.Point(3, 3))
            >>> d.is_continuous()
            False

        """
        return utils.length(self.continuous_subpaths()) <= 1

    def __continuous_subpaths(self) -> Generator[D, None, None]:
        subpath = D()

        for command in self:
            if isinstance(command, MoveTo):
                yield subpath
                subpath = D()
            else:
                subpath.append(command)

        yield subpath

    def continuous_subpaths(self) -> Iterator[D]:
        """Iterate over the continuous subpaths of the path.

        A continuous subpath is a sequence of commands that form a single
        path without any `MoveTo` commands in between.

        Returns:
            An iterator over the continuous subpaths of the path.

        Examples:
            >>> d = D()
            >>> list(d.continuous_subpaths())
            []
            >>> _ = d.move_to(point.Point(0, 0)).line_to(point.Point(1, 1))
            >>> list(d.continuous_subpaths())
            [D(LineTo(end=Point(x=1.0, y=1.0)))]
            >>> _ = d.move_to(point.Point(2, 2)).line_to(point.Point(3, 3))
            >>> for subpath in d.continuous_subpaths():
            ...     print(subpath)
            D(LineTo(end=Point(x=1.0, y=1.0)))
            D(LineTo(end=Point(x=3.0, y=3.0)))

        """
        return filter(None, self.__continuous_subpaths())

    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        commands = ", ".join(repr(command) for command in self)
        return f"{name}({commands})"

    @classmethod
    def __from_svgpathtools(cls, path: svgpathtools.Path) -> Self:
        d = cls()

        for command in path:
            match command:  # pyright: ignore[reportMatchNotExhaustive]
                case svgpathtools.Line():
                    d.line_to(point.Point.from_complex(command.end))
                case svgpathtools.QuadraticBezier():
                    d.quadratic_bezier_to(
                        control=point.Point.from_complex(command.control),
                        end=point.Point.from_complex(command.end),
                    )
                case svgpathtools.CubicBezier():
                    d.cubic_bezier_to(
                        control1=point.Point.from_complex(
                            command.control1
                        ),
                        control2=point.Point.from_complex(
                            command.control2
                        ),
                        end=point.Point.from_complex(command.end),
                    )
                case svgpathtools.Arc():
                    d.arc_to(
                        radius=point.Point.from_complex(command.radius),
                        angle=command.rotation,
                        large=command.large_arc,
                        sweep=command.sweep,
                        end=point.Point.from_complex(command.end),
                    )

        return d

    @classmethod
    def from_str(cls, value: str) -> Self:
        path = svgpathtools.parse_path(value)
        return cls.__from_svgpathtools(path)

    @override
    @classmethod
    def _validate(
        cls, value: object, info: pydantic_core.core_schema.ValidationInfo
    ) -> Self:
        del info

        match value:
            case str():
                return cls.from_str(value)
            case d if isinstance(d, cls):
                return d
            case _:
                msg = f"Expected a string or {cls.__name__}"
                raise TypeError(msg)

    def __apply_mode(self) -> Generator[PathCommand, None, None]:
        formatter = serialize.get_current_formatter()
        pos = point.Point.zero()

        for command in self:
            match formatter.path_data_mode:
                case "relative":
                    yield command - pos
                case "absolute":
                    yield command

            pos = command.end

    @staticmethod
    def __format_command(
        *args: serialize.Serializable,
        absolute_char: _AbsolutePathCommandChar,
    ) -> str:
        formatter = serialize.get_current_formatter()

        char = (
            absolute_char
            if formatter.path_data_mode == "absolute"
            else absolute_char.lower()
        )

        if not args:
            return char

        args_str = serialize.serialize(args, bool_mode="number")
        return f"{char} {args_str}"

    def __serialize_commands(self) -> Generator[str, None, None]:
        for command in self.__apply_mode():
            match command:
                case MoveTo(end):
                    yield self.__format_command(end, absolute_char="M")
                case ClosePath(end):  # has to be before LineTo
                    yield self.__format_command(absolute_char="Z")
                case LineTo(end):
                    yield self.__format_command(end, absolute_char="L")
                case QuadraticBezierTo(control, end):
                    yield self.__format_command(
                        control, end, absolute_char="Q"
                    )
                case CubicBezierTo(control1, control2, end):
                    yield self.__format_command(
                        control1, control2, end, absolute_char="C"
                    )
                case ArcTo(radius, angle, large, sweep, end):
                    yield self.__format_command(
                        radius, angle, large, sweep, end, absolute_char="A"
                    )

    @override
    def serialize(self) -> str:
        return serialize.serialize(self.__serialize_commands())


DType: TypeAlias = D



from __future__ import annotations

import re
from collections.abc import Iterable, MutableSequence
from types import TracebackType
from typing import (
    Final,
    Literal,
    Protocol,
    TypeAlias,
    overload,
    runtime_checkable,
)

import pydantic
import readable_number
from typing_extensions import TypeIs

from svglab import models


ColorSerializationMode: TypeAlias = Literal[
    "named", "hex-short", "hex-long", "rgb", "hsl", "auto", "original"
]
""" Mode for serializing colors.

- `named`: Serialize colors using their named representation, if possible.
            Falls back to `hex-short`.
- `hex-short`: Serialize colors using the short hex format
            (for example, `#fff`).
- `hex-long`: Serialize colors using the long hex format
            (for example, `#ffffff`).
- `rgb`: Serialize colors using the RGB format
            (for example, `rgb(255, 255, 255)`).
- `hsl`: Serialize colors using the HSL format
            (for example, `hsl(0, 0%, 100%)`).
- `auto`: Automatically choose the most appropriate serialization mode.
- `original`: Serialize colors using their original representation,
            if possible. Falls back to `auto`.
"""

Separator: TypeAlias = Literal[", ", ",", " "]
""" Type for basic valid separators for lists of values. """

PathDataSerializationMode: TypeAlias = Literal["relative", "absolute"]
""" Mode for serializing path data."""

BoolSerializationMode: TypeAlias = Literal["text", "number"]
""" Mode for serializing boolean values."""


@runtime_checkable
class CustomSerializable(Protocol):
    """Protocol for objects with special serialization behavior.

    When a `Serializable` object is serialized, its `serialize()` method is
    used to obtain the string representation, instead of using `str()`.
    """

    def serialize(self) -> str:
        """Return an SVG-friendly string representation of this object."""
        ...


Serializable: TypeAlias = (
    bool
    | int
    | float
    | str
    | CustomSerializable
    | Iterable["Serializable"]
)
""" Type for objects that can be serialized to a SVG-friendly string. """


def _is_serializable(value: object, /) -> TypeIs[Serializable]:
    return isinstance(
        value,
        bool
        | int
        | float
        | str
        | MutableSequence
        | tuple
        | CustomSerializable,
    )


class Formatter(models.BaseModel):
    """Formatter for serializing SVG elements.

    This class, together with `set_formatter()` and `get_current_formatter()`,
    can be used to customize the serialization of SVG elements.

    Attributes:
    `color_mode`: The color serialization mode (`hsl`, `rgb`, ...)
    to use when serializing colors.

    `show_decimal_part_if_int`: Whether to show the decimal part of a number
    even if it is an integer. For example, `1.0` instead of `1`.
    `max_precision`: The maximum number of significant digits
    after the decimal point to use when serializing numbers.
    `small_number_scientific_threshold`: The magnitude threshold below which
    numbers are serialized using scientific notation.
    For example, `1e-06` instead of `0.000001`. If `None`, scientific notation
    is not used for small numbers.
    `large_number_scientific_threshold`: The magnitude threshold above which
    numbers are serialized using scientific notation.
    For example, `1e+06` instead of `1000000`. If `None`, scientific notation
    is not used for large numbers.

    `path_data_mode`: The path data serialization mode (`relative`, `absolute`)

    `list_separator`: The separator to use when serializing lists of values.
    `point_separator`: The separator to use when serializing points.

    `indent`: The number of spaces to use for indentation in the resulting
    SVG document.
    `spaces_around_attrs`: Whether to add spaces around attribute values.
    For example, `fill=" red "` instead of `fill="red"`.
    `spaces_around_function_args`: Whether to add spaces around function
    arguments. For example, `rotate( 45 )` instead of `rotate(45)`.

    """

    model_config = pydantic.ConfigDict(frozen=True)

    # colors
    color_mode: models.KwOnly[ColorSerializationMode] = "auto"

    # numbers
    show_decimal_part_if_int: models.KwOnly[bool] = False
    max_precision: models.KwOnly[int | None] = pydantic.Field(
        default=None, ge=0
    )
    small_number_scientific_threshold: models.KwOnly[float | None] = (
        pydantic.Field(default=1e-6, ge=0)
    )
    large_number_scientific_threshold: models.KwOnly[int | None] = (
        pydantic.Field(default=int(1e6), ge=0)
    )

    # path data
    path_data_mode: models.KwOnly[PathDataSerializationMode] = "absolute"

    # separators
    list_separator: models.KwOnly[Separator] = " "
    point_separator: models.KwOnly[Separator] = ","

    # whitespace
    indent: models.KwOnly[int] = pydantic.Field(default=2, ge=0)
    spaces_around_attrs: models.KwOnly[bool] = False
    spaces_around_function_args: models.KwOnly[bool] = False

    __original_formatter: Formatter | None = pydantic.PrivateAttr(
        default=None
    )

    def __enter__(self) -> None:
        self.__original_formatter = get_current_formatter()
        set_formatter(self)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        del exc_type, exc_val, exc_tb

        if self.__original_formatter is None:
            msg = (
                "Cannot exit unentered context manager.",
                " Call `__enter__` first",
            )

            raise RuntimeError(msg)

        set_formatter(self.__original_formatter)


DEFAULT_FORMATTER: Final = Formatter()
"""The default formatter used by the library.
Use `set_formatter()` to set a custom one."""

_formatter = DEFAULT_FORMATTER


def get_current_formatter() -> Formatter:
    """Obtain a reference to the current formatter."""
    return _formatter


def set_formatter(formatter: Formatter) -> None:
    """Set the current formatter."""
    global _formatter  # noqa: PLW0603
    _formatter = formatter


@overload
def _serialize_number(number: float, /) -> str: ...


@overload
def _serialize_number(
    first: float, second: float, /, *numbers: float
) -> tuple[str, ...]: ...


def _serialize_number(*numbers: float) -> str | tuple[str, ...]:
    """Format a number or a sequence of numbers for SVG serialization.

    Args:
    numbers: The numbers to format.

    Returns:
    The formatted number or numbers.

    Examples:
    >>> _serialize_number(42)
    '42'
    >>> _serialize_number(3.14)
    '3.14'
    >>> _serialize_number(1, 2, 3)
    ('1', '2', '3')
    >>> _serialize_number(1.0, 2.0, 3.0)
    ('1', '2', '3')
    >>> _serialize_number(1e9)
    '1e+09'

    """
    formatter = get_current_formatter()

    rn = readable_number.ReadableNumber(
        digit_group_delimiter="",  # group separators are not allowed in SVG
        significant_figures_after_decimal_point=formatter.max_precision,
        show_decimal_part_if_integer=formatter.show_decimal_part_if_int,
        use_exponent_for_small_numbers=(
            formatter.small_number_scientific_threshold is not None
        ),
        small_number_threshold=formatter.small_number_scientific_threshold
        or 0,
        use_exponent_for_large_numbers=(
            formatter.large_number_scientific_threshold is not None
        ),
        large_number_threshold=formatter.large_number_scientific_threshold
        or 0,
    )

    result = tuple(rn.of(number) for number in numbers)

    return result[0] if len(result) == 1 else result


def serialize_attr(value: object, /) -> str:
    """Serialize an attribute value into its SVG representation.

    Args:
    value: The value to serialize.

    Returns:
    The SVG representation of the value.

    Examples:
    >>> serialize_attr(42)
    '42'
    >>> serialize_attr(3.14)
    '3.14'
    >>> serialize_attr("foo")
    'foo'
    >>> serialize_attr(["foo", "bar"])
    'foo bar'

    """
    if not _is_serializable(value):
        msg = f"Type {type(value)} is not serializable."
        raise TypeError(msg)

    formatter = get_current_formatter()
    result = serialize(value)

    if formatter.spaces_around_attrs:
        result = f" {result} "

    return result


def _extract_function_name_and_args(attr: str) -> tuple[str, str] | None:
    """Extract function name and arguments from a function-call-like attribute.

    An attribute is considered to be a function call if it has the form
    `name(args)`. This function extracts the name and the arguments from such
    an attribute. If the attribute is not a function call, `None` is returned.

    Args:
    attr: The attribute to extract the function name and arguments from.

    Returns:
    A tuple containing the function name and the arguments,
    or `None` if the attribute is not a function call.

    Examples:
    >>> _extract_function_name_and_args("foo()") is None  # no arguments
    True
    >>> _extract_function_name_and_args("foo(42)")
    ('foo', '42')
    >>> _extract_function_name_and_args("foo(42, 'bar')")
    ('foo', "42, 'bar'")
    >>> _extract_function_name_and_args(
    ...     "bar"
    ... ) is None  # not a function call
    True

    """
    match = re.match(r"^([^\(\)]+)\(([^\(\)]+)\)$", attr)

    if match is None:
        return None

    return match.group(1), match.group(2)


@overload
def serialize(
    value: Serializable, /, *, bool_mode: BoolSerializationMode = "text"
) -> str: ...


@overload
def serialize(
    first: Serializable,
    second: Serializable,
    /,
    *values: Serializable,
    bool_mode: BoolSerializationMode = "text",
) -> tuple[str, ...]: ...


def serialize(
    *values: Serializable, bool_mode: BoolSerializationMode = "text"
) -> str | tuple[str, ...]:
    """Return an SVG-friendly string representation of the given value(s)."""
    result = tuple(
        _serialize(value, bool_mode=bool_mode) for value in values
    )
    return result[0] if len(result) == 1 else result


def _serialize(
    value: Serializable, /, *, bool_mode: BoolSerializationMode
) -> str:
    formatter = get_current_formatter()
    result: str

    match value:
        case CustomSerializable():
            result = value.serialize()

            if (
                formatter.spaces_around_function_args
                and (fn_call := _extract_function_name_and_args(result))
                is not None
            ):
                fn, args = fn_call
                result = f"{fn}( {args} )"
        # needs to be before int (bool is a subclass of int)
        case bool():
            result = _serialize_bool(value, mode=bool_mode)
        case int() | float():
            result = _serialize_number(value)
        case str():
            result = value
        case bytes():
            result = value.decode()
        # this should go last to avoid classifying strings as iterables, etc.
        case Iterable():
            result = formatter.list_separator.join(
                _serialize(v, bool_mode=bool_mode) for v in value
            )

    return result


def _serialize_bool(
    value: bool,  # noqa: FBT001
    /,
    *,
    mode: BoolSerializationMode,
) -> str:
    """Serialize a boolean value into its SVG representation.

    Args:
    value: The boolean value to serialize.
    mode: The serialization mode to use. Can be either "text" or "number".
    If set to "text", the value is serialized as "true" or "false".
    If set to "number", the value is serialized as "1" or "0".

    Returns:
    The SVG representation of the boolean value.

    Examples:
    >>> _serialize_bool(True, mode="text")
    'true'
    >>> _serialize_bool(False, mode="text")
    'false'
    >>> _serialize_bool(True, mode="number")
    '1'
    >>> _serialize_bool(False, mode="number")
    '0'

    """
    match mode:
        case "text":
            return "true" if value else "false"
        case "number":
            return "1" if value else "0"
